//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                         Floppy disk support module                         *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// This module implements functions to represent the actual physical disks.
// It consists of support functions that manipulate the disk image files.
//
// To do:
//    - Check that seek position is within range of image.  If not then put
//      message out and flag as an error.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.1.0 - 14 August 2007, S.J.Kay
// - Added Microbee ds40, ss80, ds80, ds82, ds84 raw disk image types.
// - Modified DSK code to detect all Microbee formats.  This needs testing for
//   some formats. (warning in README)
//
// v1.0.0 - 19 June 2007, S.J.Kay
// The following changes/additions have been made:
// - Added new disk image structures,  this now allows for different floppy
//   images to be used instead of only the 40T DS (SBC) images.
// - disk_open function now uses open with write enabled (r+b). This is
//   required to allow disk_write to work.
// - Added an assert to test max_size variable in disk_read function.
// - Implement the disk_write function.
//
// v0.0.0 - 7 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>

#include "disk.h"
#include "ubee512.h"

//==============================================================================
// Disk open.
//
// Determine the disk format and structure being used.  This can one of the
// following:
//
// * original v0.22 raw disk format (40T DS DD) Side0=1st 200K, Side1=2nd 200K
// * dsk format
// * dip format (disk image plus) This is an in-house spec not finalised.
// * ds40 - Microbee 40T DS DD (SBC) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ss80 - Microbee 80T SS DD (CIAB) S0-T0, S0-T1, S0-T2 ...
// * ds80 - Microbee 80T DS DD (modular) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ds82 - Microbee 80T DS DD (dream) S0-T0, S1-T0, S0-T1, S1-T1 ...
// * ds84 - Microbee 80T DS DD (pjb) S0-T0, S1-T0, S0-T1, S1-T1 ...
//
//   pass: const char *name
//         disk_t *disk
// return: int				0 if no error, else -1 
//==============================================================================
int disk_open (const char *name, disk_t *disk)
{
 dski_t dski;			// dsk information block
 dskt_t dskt;			// dsk track information block

 int trackofs;
 int dskofs; 

 char uname[512];
 
 char const image_types[][5] =
 {
  ".DIP",
  ".DSK",
  ".IMG",
  ".NW",
  ".DS40",
  ".SS80",
  ".DS80",
  ".DS82",
  ".DS84",
  ""
 };
 
 int i = 0;
 uname[0] = 0; 
 strcat(uname, name);
 while (uname[i])
    {
     uname[i] = toupper(uname[i]);
     i++;
    }

 assert(disk != NULL);

 disk->fdisk = fopen(name, "r+b");

 if (disk->fdisk == NULL)
    return -1;

 for (i=0; image_types[i][0] != 0; i++)
    {
     if (strstr(uname, image_types[i]) != NULL)
        break;
    }
 if (image_types[i][0] == 0)
    return -1;    
 disk->itype = i;
 
 disk->wrprot = 0;

 switch (disk->itype)
    {
     case 0 : // DIP (in house spec)
              fseek(disk->fdisk, -sizeof(disk->imagerec), SEEK_END);
              fread(&disk->imagerec, sizeof(disk->imagerec), 1, disk->fdisk);
              if (strcmp(disk->imagerec.id, "DISK-IMAGE-PLUS") != 0)
                 return -1;
              if ((disk->imagerec.secsize != 128) && (disk->imagerec.secsize != 256) &&
                 (disk->imagerec.secsize != 512) && (disk->imagerec.secsize != 1024))
                 return -1;
              disk->wrprot = disk->imagerec.wrprot;
#ifdef use_info_disk_open
              printf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
              disk->imagerec.ver, disk->imagerec.type);
#endif
              break;
     case 1 : // DSK
// get the disk information block
              fread(&dski, sizeof(dski), 1, disk->fdisk);
              disk->imagerec.tracks = dski.tracks;
              disk->imagerec.heads = dski.heads; 

              if (dski.heads == 2)
                 disk->imagerec.datatrack = 1;
              else   
                 disk->imagerec.datatrack = 2;

// get the first track information block to get some initial values
              fread(&dskt, sizeof(dskt), 1, disk->fdisk);
              disk->imagerec.sectrack = dskt.spt;
              disk->imagerec.secsize = 128 << dskt.bps;
              
// get a sample from track 5 side 0,  we want to get the first sector
// number from a data track to see if this is Microbee Modular disk
              trackofs = 5 * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              dskofs = (((5 * disk->imagerec.heads) + 1) * 0x100) + 0x100;
              fseek(disk->fdisk, trackofs + dskofs, SEEK_SET);              
              fread(&dskt, sizeof(dskt), 1, disk->fdisk);

// not sure if this will work for the Microbee Modular disk format (warning is in README)              
// STOP PRESS - Let's not use it for now until we get it sorted (if even possible)
#if 0
              if (dskt.sect_numb >= 20)                 
                 disk->imagerec.datasecofs = 21;
              else   
                 disk->imagerec.datasecofs = 1;
              disk->imagerec.systsecofs = 1;                 
              printf("dskt.sect_numb=%d\n", dskt.sect_numb);              
              printf("dskt.hnumb_sectid=%d\n",  dskt.hnumb_sectid);
#else
              disk->imagerec.datasecofs = 1;
              disk->imagerec.systsecofs = 1;                 
#endif   

#ifdef use_info_disk_open
              printf("Image format is DSK\n");
#endif
              break;
     case 2 : // IMG (nanowasp v0.22 format)
     case 3 : // NW (nanowasp v0.22 format)
              disk->imagerec.tracks = 40;
              disk->imagerec.heads = 2;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           

#ifdef use_info_disk_open
              printf("Image format is RAW 40T DS DD\n");
#endif
              break;
     case 4 : // DS40
              disk->imagerec.tracks = 40;
              disk->imagerec.heads = 2;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           
              disk->imagerec.datatrack = 1;
              disk->imagerec.datasecofs = 1;
              disk->imagerec.systsecofs = 1;
#ifdef use_info_disk_open
              printf("Image format is RAW DS40\n");
#endif
              break;
     case 5 : // SS80
              disk->imagerec.tracks = 80;
              disk->imagerec.heads = 1;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           
              disk->imagerec.datatrack = 2;
              disk->imagerec.datasecofs = 1;
              disk->imagerec.systsecofs = 1;

#ifdef use_info_disk_open
              printf("Image format is RAW SS80\n");
#endif
              break;

     case 6 : // DS80 (Microbee modular)
              disk->imagerec.tracks = 80;
              disk->imagerec.heads = 2;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           
              disk->imagerec.datatrack = 1;
              disk->imagerec.datasecofs = 21;	// sectors are 21 to 30
              disk->imagerec.systsecofs = 1;
#ifdef use_info_disk_open
              printf("Image format is RAW DS80\n");
#endif
              break;
     case 7 : // DS82
              disk->imagerec.tracks = 80;
              disk->imagerec.heads = 2;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           
              disk->imagerec.datatrack = 1;
              disk->imagerec.datasecofs = 1;
              disk->imagerec.systsecofs = 1;
#ifdef use_info_disk_open
              printf("Image format is RAW DS82\n");
#endif
              break;
     case 8 : // DS84
              disk->imagerec.tracks = 80;
              disk->imagerec.heads = 2;
              disk->imagerec.sectrack = 10;
              disk->imagerec.secsize = 512;           
              disk->imagerec.datatrack = 1;
              disk->imagerec.datasecofs = 1;
              disk->imagerec.systsecofs = 1;
#ifdef use_info_disk_open
              printf("Image format is RAW DS84\n");
#endif
              break;
    } 

#ifdef use_info_disk_open
     printf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
     disk->imagerec.tracks, disk->imagerec.heads,
     disk->imagerec.sectrack, disk->imagerec.secsize);
#endif

 return 0;    
}

//==============================================================================
// Disk close.
//
//   pass: disk_t *disk
// return: void
//==============================================================================
void disk_close(disk_t *disk)
{
 assert(disk != NULL);

 fclose(disk->fdisk);
 disk->fdisk = NULL;
}

//==============================================================================
// Disk create.
//
//   pass: const char *name
//         disk_t *disk
// return: int				-1
//==============================================================================
int disk_create(const char *name, disk_t *disk)
{
 return -1;
}

//==============================================================================
// Disk read.
//
//   pass: disk_t *disk
//         char *buf
//         int max_size
//         int side
//         int track
//         int sect
// return: int				1 
//==============================================================================
int disk_read(disk_t *disk, char *buf, int max_size, int side, int track, int sect)
{
 int trackofs;
 int sectofs;
 int sectuse;
 int dskofs;

 assert(disk != NULL);
 assert(disk->fdisk != NULL);

#ifdef use_debug_disk_read_abort
 printf("disk_read: forcing an abort here\n");
 assert(1 != 1);
#endif

 switch (disk->itype)
    {
     case 0 : // DIP (in house spec)
              if (track >= disk->imagerec.datatrack)
                 sectuse = sect - disk->imagerec.datasecofs;
              else
                 sectuse = sect - disk->imagerec.systsecofs;

              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);

              fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
              fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
#ifdef use_info_disk_read
              printf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
              disk->imagerec.ver, disk->imagerec.type);
#endif
              break;
     case 1 : // DSK
              if (track >= disk->imagerec.datatrack)
                 sectuse = sect - disk->imagerec.datasecofs;
              else
                 sectuse = sect - disk->imagerec.systsecofs;

              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
     
              dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;

              fseek(disk->fdisk, trackofs + sectofs + dskofs, SEEK_SET);
              fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
              break;
     case 2 : // IMG (nanowasp v0.22 format)
     case 3 : // NW (nanowasp v0.22 format)
              fseek(disk->fdisk, ((side * 40 + track) * 10 + (sect - 1)) * 512, SEEK_SET);
              fread(buf, 512, 1, disk->fdisk);
              break;
     case 4 : // DS40
     case 5 : // SS80
     case 6 : // DS80 (Microbee modular)
     case 7 : // DS82
     case 8 : // DS84
              if (track >= disk->imagerec.datatrack)
                 sectuse = sect - disk->imagerec.datasecofs;
              else
                 sectuse = sect - disk->imagerec.systsecofs;
              
              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);

              fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);
              fread(buf, disk->imagerec.secsize, 1, disk->fdisk);
              break;
    }  

#ifdef use_info_disk_read
 printf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
 disk->imagerec.tracks, disk->imagerec.heads,
 disk->imagerec.sectrack, disk->imagerec.secsize);
 printf("Trk: %d   Sde: %d   Sec: %d\n",  track, side, sect);
 if (disk->itype == 1)
    printf("File offset: %d\n\n", trackofs + sectofs + dskofs);
 else   
    printf("File offset: %d\n\n", trackofs + sectofs);
#endif
 
 return 1;
}

//==============================================================================
// Disk write.
//
// A flush call is used on each sector write.  This may be moved out at a later
// stage when image (floppy) swapping is implemented.
//
//   pass: disk_t *disk
//         char *buf
//         int size
//         int side
//         int track
//         int sect
// return: int				1 
//==============================================================================
int disk_write(disk_t *disk, char *buf, int size, int side, int track, int sect)
{
 int trackofs;
 int sectofs;
 int sectuse; 
 int dskofs;
 
 assert(disk != NULL);
 assert(disk->fdisk != NULL);

 switch (disk->itype)
    {
     case 0 : // DIP (in house spec)
              if (track >= disk->imagerec.datatrack)
                 sectuse = sect - disk->imagerec.datasecofs;
              else
                 sectuse = sect - disk->imagerec.systsecofs;       

              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
     
              fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);

              fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
              fflush(disk->fdisk);
#ifdef use_info_disk_read
              printf("Image: %s   Vers: %s   Type: %s\n", disk->imagerec.id,
              disk->imagerec.ver, disk->imagerec.type);
#endif
              break;
     case 1 : // DSK
              if (track >= disk->imagerec.datatrack)
                 sectuse = sect - disk->imagerec.datasecofs;
              else
                 sectuse = sect - disk->imagerec.systsecofs;

              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
     
              dskofs = (((track * disk->imagerec.heads) + side + 1) * 0x100) + 0x100;
     
              fseek(disk->fdisk, trackofs + sectofs + dskofs, SEEK_SET);

              fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
              fflush(disk->fdisk);
              break;
     case 2 : // IMG (nanowasp v0.22 format)
     case 3 : // NW (nanowasp v0.22 format)
              fseek(disk->fdisk, ((side * 40 + track) * 10 + (sect - 1)) * 512, SEEK_SET);
              fwrite(buf, 512, 1, disk->fdisk);
              fflush(disk->fdisk);
              break;
     case 4 : // DS40
     case 5 : // SS80
     case 6 : // DS80 (Microbee modular)
     case 7 : // DS82
     case 8 : // DS84
              if (track >= disk->imagerec.datatrack)
                 sectuse = sect - disk->imagerec.datasecofs;
              else
                 sectuse = sect - disk->imagerec.systsecofs;

              trackofs = track * disk->imagerec.heads * disk->imagerec.sectrack * disk->imagerec.secsize;
              sectofs = side * disk->imagerec.sectrack * disk->imagerec.secsize + (sectuse * disk->imagerec.secsize);
     
              fseek(disk->fdisk, trackofs + sectofs, SEEK_SET);

              fwrite(buf, disk->imagerec.secsize, 1, disk->fdisk);
              fflush(disk->fdisk);
              break;
    }  

#ifdef use_info_disk_read
 printf("Tracks: %d   Heads: %d   S/T: %d   B/S: %d\n", 
 disk->imagerec.tracks, disk->imagerec.heads,
 disk->imagerec.sectrack, disk->imagerec.secsize);
 printf("Trk: %d   Sde: %d   Sec: %d\n",  track, side, sect);
 if (disk->itype == 1)
    printf("File offset: %d\n\n", trackofs + sectofs + dskofs);
 else   
    printf("File offset: %d\n\n", trackofs + sectofs);

#endif
    
 return 1;    
}

//==============================================================================
// disk get sector header ???.
//
//   pass:
// return:
//==============================================================================

/* int disk_getsecthdr(...) ?? */

//==============================================================================
// Disk write protected query.
//
// To add:  Disk images could have an '_' underscore character appended to the
// end of the filename to make the images write protected.  This would mean that
// system dependent permissions would not be require and would be portable.
//
//   pass: disk_t *disk
// return: int				write protect status
//==============================================================================
int disk_iswrprot(disk_t *disk)
{
   if (disk == NULL)
      return 0;

   return disk->wrprot;
}
