//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                              WD2793 FDC module                             *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Emulates the WD2793 floppy disk controller.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.3.0 - 1 September 2007, S.J.Kay
// - Changes some error messages and redirection.
//
// v1.0.0 - 2 July 2007, S.J.Kay
// - Implement DSK disk image format instead of raw images.
// - All references to '10' sectors per track have been replaced by
//   drive->disk->imagerec.sectrack and references to '512' have been replaced
//   by drive->disk->imagerec.secsize.
// - Disk images are no longer assumed to MAXTRACK=39,  for now change to
//   255 (largest value allowed for WD2793 ?)
// - BUFSSZE increased from 10x512 to 10x1024 byte sectors.
// - Added 2 more drives (now is A-D).  The drive images can be specified on
//   the command line. If drive 'A' is not specified then "boot.dsk" will be
//   used by default.  Drives 'B-D' must be specified if required.
// - implemented case FDC_WRITESECT: in fdc_cmd function.
// - implemented fdc_data_w function.
// - in fdc_data_r changed the ctrl_status value at end sector read.  The
//   value of FDC_WRPROT did not look correct here.
// - Added a second drive.  
//
// v0.0.0 - 5 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>

#include "fdc.h"
#include "disk.h"
#include "ubee512.h"

#define FDC_RESTORE      0       /* type I */
#define FDC_SEEK         1
#define FDC_STEP         2
#define FDC_STEPIN       4
#define FDC_STEPOUT      6

#define FDC_READSECT     8       /* type II */
#define FDC_WRITESECT    10

#define FDC_READADDR     12      /* type III */
#define FDC_READTRACK    14
#define FDC_WRITETRACK   15

#define FDC_INTERRUPT    13      /* type IV */


#define FDC_NOTREADY     0x80    /* status port bits */
#define FDC_WRPROT       0x40
#define FDC_HEADLOADED   0x20
#define FDC_RECTYPE      0x20
#define FDC_SEEKERROR    0x10
#define FDC_RECNOTFOUND  0x10
#define FDC_CRCERROR     0x08
#define FDC_TRACK0       0x04
#define FDC_LOSTDATA     0x04
#define FDC_INDEXPULSE   0x02
#define FDC_DRQ          0x02
#define FDC_BUSY         0x01

#define FDC_STEPRATE     0x03    /* fdc command data bits */
#define FDC_VERIFY       0x04
#define FDC_LOADHEAD     0x08
#define FDC_UPDATETRACK  0x10
#define FDC_MULTISECT    0x10
#define FDC_SIDE         0x08
#define FDC_DELAY        0x04
#define FDC_CMPSIDE      0x02
#define FDC_DATAMARK     0x01
#define FDC_INTREADY     0x01
#define FDC_INTNOTREADY  0x02
#define FDC_INTINDEX     0x04
#define FDC_INTIMMED     0x08

#define FDC_MAXTRACK     255

#define FDC_BUFSIZE      1024*10	// maximum data per track HD Floppy

//==============================================================================
// structures and variables
//==============================================================================
struct fdc_drive_s
{
 int track;                             /* current track position of the head */
 disk_t *disk;
};

typedef struct fdc_drive_s fdc_drive_t;

static int ctrl_side, ctrl_drive, ctrl_ddense;
static int ctrl_rdata, ctrl_rtrack, ctrl_rsect,
           ctrl_intrq, ctrl_drq, ctrl_status,
           ctrl_stepdir;


static fdc_drive_t fdc_drive[FDC_NUMDRIVES];

static char buf[FDC_BUFSIZE];
static int bytes_left, buf_index;

extern char floppya[];
extern char floppyb[];
extern char floppyc[];
extern char floppyd[];

extern FILE *debuglog;
extern int debuglvl;

//==============================================================================
// Initialise
//
//   pass: void
// return: int				0 if no errors, else -1
//==============================================================================
int fdc_init(void)
{
 int i;

 for (i=0; i < FDC_NUMDRIVES; i++)
    fdc_drive[i].disk = NULL;

 if (strlen(floppya))
    {
     if (fdc_loaddisk(floppya, 0))
        return -1;
    }    
 else 
    return -1;

 if (strlen(floppyb))
    if (fdc_loaddisk(floppyb, 1))
       return -1;    

 if (strlen(floppyc))
    if (fdc_loaddisk(floppyc, 2))
       return -1;

 if (strlen(floppyd))
    if (fdc_loaddisk(floppyd, 3))
       return -1;

 return 0;
}

//==============================================================================
// De-initialise
//
//   pass: void
// return: int				0
//==============================================================================
int fdc_deinit(void)
{
 int i;

 for (i=0; i<FDC_NUMDRIVES; i++)
    fdc_unloaddisk(i);
 
 return 0;
}

//==============================================================================
// reset the controller
//
//   pass: void
// return: int				0
//==============================================================================
int fdc_reset(void)
{
 ctrl_side = ctrl_drive = ctrl_ddense = 0;
 ctrl_rtrack = ctrl_intrq = ctrl_drq = 0;
 ctrl_stepdir = +1;

 fdc_drive[ctrl_drive].track = 0;
 ctrl_status = FDC_TRACK0;

 if (fdc_drive[ctrl_drive].disk == NULL)
    ctrl_status |= FDC_INDEXPULSE;              /* set index pulse if no disk */
 if (disk_iswrprot(fdc_drive[ctrl_drive].disk))
    ctrl_status |= FDC_WRPROT;
 
 bytes_left = 0;

 return 0;
}

//==============================================================================
// load disk
//
//   pass: const char *name
//         int drive
// return: int				0 if no errors,  else -1
//==============================================================================
int fdc_loaddisk(const char *name, int drive)
{
 if (fdc_drive[drive].disk != NULL)
    fdc_unloaddisk(drive);

 fdc_drive[drive].disk = malloc(sizeof(disk_t));
 
 if (fdc_drive[drive].disk == NULL)
    {
     printf("fdc_loaddisk: Load disk error");
     exit(1);
    }

 if (disk_open(name, fdc_drive[drive].disk) != 0)
    {
     printf("fdc_loaddisk: Unable to load %s\n", name);
     return -1;
    }
 
 return 0;
}

//==============================================================================
// un-load disk
//
//   pass: int drive			drive number
// return: void
//==============================================================================
void fdc_unloaddisk(int drive)
{
 if (fdc_drive[drive].disk == NULL)
    return;

 disk_close(fdc_drive[drive].disk);
 free(fdc_drive[drive].disk);
 fdc_drive[drive].disk = NULL;
}

//==============================================================================
// FDC command
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_cmd(UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 int cmd;
 fdc_drive_t *drive = &fdc_drive[ctrl_drive];

 ctrl_intrq = 0;                      /* new command clears interrupt, FIXME for
                                                     immediate interrupt case */

 cmd = (data >> 4) & 0x0F;                                /* get command bits */
 if ( ((cmd >> 1) >= 1) && ((cmd >> 1) <= 5) )
    cmd &= 0x0E;                 /* get rid of data bits in the command field */

 switch (cmd)
    {
     case FDC_RESTORE:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: restore\n");

        drive->track = ctrl_rtrack = 0;

        ctrl_intrq = 1;
        ctrl_status = FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(drive->disk))
           ctrl_status |= FDC_WRPROT;
        if (drive->disk == NULL)
           ctrl_status |= FDC_INDEXPULSE;
        break;

     case FDC_SEEK:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: seek\n");

        drive->track += ctrl_rdata - ctrl_rtrack;
        if (drive->track < 0)
           drive->track = 0;
        else if (drive->track > FDC_MAXTRACK)
           drive->track = FDC_MAXTRACK;

        if (ctrl_rdata < ctrl_rtrack)
           ctrl_stepdir = -1;
        else
           ctrl_stepdir = +1;
        ctrl_rtrack = ctrl_rdata;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (drive->track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(drive->disk))
           ctrl_status |= FDC_WRPROT;
        if (drive->disk == NULL)
           ctrl_status |= FDC_INDEXPULSE;
        break;

     case FDC_STEP:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: step\n");

        drive->track += ctrl_stepdir;
        if (drive->track < 0)
           drive->track = 0;
        else
           if (drive->track > FDC_MAXTRACK)
              drive->track = FDC_MAXTRACK;

        if (data & FDC_UPDATETRACK)
           ctrl_rtrack += ctrl_stepdir;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (drive->track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(drive->disk))
           ctrl_status |= FDC_WRPROT;
        if (drive->disk == NULL)
           ctrl_status |= FDC_INDEXPULSE;
        break;

     case FDC_STEPIN:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: stepin\n");

        drive->track++;
        if (drive->track > FDC_MAXTRACK)
           drive->track = FDC_MAXTRACK;

        if (data & FDC_UPDATETRACK)
           ctrl_rtrack++;

        ctrl_stepdir = +1;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (drive->track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(drive->disk))
           ctrl_status |= FDC_WRPROT;
        if (drive->disk == NULL)
           ctrl_status |= FDC_INDEXPULSE;
        break;

     case FDC_STEPOUT:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: stepout\n");

        drive->track--;
        if (drive->track < 0)
           drive->track = 0;

        if (data & FDC_UPDATETRACK)
           ctrl_rtrack--;

        ctrl_stepdir = -1;

        ctrl_intrq = 1;
        ctrl_status = 0;
        if (drive->track == 0)
           ctrl_status |= FDC_TRACK0;
        if (data & (FDC_LOADHEAD | FDC_VERIFY))
           ctrl_status |= FDC_HEADLOADED;
        if (disk_iswrprot(drive->disk))
           ctrl_status |= FDC_WRPROT;
        if (drive->disk == NULL)
           ctrl_status |= FDC_INDEXPULSE;
        break;

     case FDC_READSECT:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: readsect\n");

        if ((drive->disk == NULL) || (ctrl_rsect < 1) || (ctrl_rsect > drive->disk->imagerec.sectrack))
           {
            ctrl_status = FDC_RECNOTFOUND;
            break;
           }

        disk_read(drive->disk, buf, drive->disk->imagerec.secsize, ctrl_side, drive->track, ctrl_rsect);        
        buf_index = 0;
        bytes_left = drive->disk->imagerec.secsize;        

        if (data & FDC_MULTISECT)
           while (++ctrl_rsect <= drive->disk->imagerec.sectrack)           
              {
               disk_read(drive->disk, &buf[bytes_left], drive->disk->imagerec.secsize, ctrl_side, drive->track, ctrl_rsect);               
               bytes_left += drive->disk->imagerec.secsize;               
              }

        ctrl_status = FDC_BUSY | FDC_DRQ;
        ctrl_drq = 1;
        break;

     case FDC_WRITESECT:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: writesect\n");

        if ((drive->disk == NULL) || (ctrl_rsect < 1) || (ctrl_rsect > drive->disk->imagerec.sectrack))
           {
            ctrl_status = FDC_RECNOTFOUND;
            break;
           }

        buf_index = 0;
        bytes_left = drive->disk->imagerec.secsize;        

        if (data & FDC_MULTISECT)
           while (++ctrl_rsect <= drive->disk->imagerec.sectrack)           
               bytes_left += drive->disk->imagerec.secsize;               

        ctrl_status = FDC_BUSY | FDC_DRQ;
        ctrl_drq = 1;
        break;

     case FDC_READADDR:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: readaddr\n");

        bytes_left = 6;
        buf_index = 0;
        buf[0] = ctrl_rsect = drive->track;
        buf[1] = ctrl_side;
        buf[2] = 1;
        buf[3] = 2;
        buf[4] = buf[5] = 0xFF;                                /* crc, FIXME! */
        ctrl_status = FDC_BUSY | FDC_DRQ;
        ctrl_drq = 1;
        break;

     case FDC_READTRACK:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: readtrack\n");
        break;

     case FDC_WRITETRACK:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: writetrack\n");
        break;

     case FDC_INTERRUPT:
        if (debuglvl > 0)
           fprintf(debuglog, "fdc_cmd: interrupt\n");

        ctrl_status = FDC_WRPROT;                        /* FIXME: check this */
        bytes_left = 0;
        ctrl_drq = 0;
        ctrl_intrq = 1;
        break;
    }
}

//==============================================================================
// Read FDC status
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			ctrl_status
//==============================================================================
UINT16 fdc_status(UINT16 port, struct z80PortRead *port_s)
{
 ctrl_intrq = 0;                     /* read status clears interrupt, FIXME for
                                                    immediate interrupt case */

 if (bytes_left != 0)                                       /* FIXME: timing */
    {
     buf_index++;
     ctrl_status |= FDC_LOSTDATA;
     if (--bytes_left == 0)
        {
         ctrl_status = FDC_WRPROT | FDC_LOSTDATA;                  /* FIXME! */
         ctrl_drq = 0;
         ctrl_intrq = 1;
        }
    }
 
 if (debuglvl > 0)
    fprintf(debuglog, "fdc_status: 0x%02X\n", ctrl_status); 
    
// printf("fdc_status: 0x%02X\n", ctrl_status);    
 return ctrl_status;
}

//==============================================================================
// Write 1 data byte to sector
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_data_w(UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 fdc_drive_t *drive = &fdc_drive[ctrl_drive];

 if (bytes_left != 0)
    {
     buf[buf_index++] = data;
     if (--bytes_left == 0)            	// if end of sector write
        {
         disk_write(drive->disk, buf, drive->disk->imagerec.secsize, ctrl_side, drive->track, ctrl_rsect);         
         ctrl_drq = 0;
         ctrl_intrq = 1;
         ctrl_status &= 0xFF ^ (FDC_BUSY | FDC_DRQ );
        }
    }

 if (debuglvl > 0)
    fprintf(debuglog, "fdc_data_w: data = 0x%02X\n", data);

 ctrl_rdata = data;
}

//==============================================================================
// Read 1 data byte from sector
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: UINT16 			ctrl_rdata
//==============================================================================
UINT16 fdc_data_r(UINT16 port, struct z80PortRead *port_s)
{
 if (bytes_left != 0)
    {
     ctrl_rdata = buf[buf_index++];
     if (--bytes_left == 0)		// if end of sector read
        {
         ctrl_drq = 0;
         ctrl_intrq = 1;
         ctrl_status &= 0xFF ^ (FDC_BUSY | FDC_DRQ );         
        }
    }

 if (debuglvl > 0)
    fprintf(debuglog, "fdc_data_r: data = 0x%02X\n", ctrl_rdata & 0xFF);

 return ctrl_rdata & 0xFF;
}

//==============================================================================
// Set track number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_track_w(UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (debuglvl > 0)
    fprintf(debuglog, "fdc_track_w: called\n");
#if 0
 printf("fdc_track_w: %d\n", data);    
#endif
 ctrl_rtrack = data;
}

//==============================================================================
// Get track number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: UINT16			ctrl_rtrack
//==============================================================================
UINT16 fdc_track_r(UINT16 port, struct z80PortRead *port_s)
{
 if (debuglvl > 0)
    fprintf(debuglog, "fdc_track_r: called\n");
#if 0
 printf("fdc_track_r: %d\n", ctrl_rtrack);    
#endif
 return ctrl_rtrack;
}

//==============================================================================
// Set sector number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_sect_w(UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (debuglvl > 0)
    fprintf(debuglog, "fdc_sect_w: called\n");
 ctrl_rsect = data;
}

//==============================================================================
// Get sector number
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: UINT16			ctrl_rsect
//==============================================================================
UINT16 fdc_sect_r(UINT16 port, struct z80PortRead *port_s)
{
 if (debuglvl > 0)
    fprintf(debuglog, "fdc_sect_r: called\n");
 return ctrl_rsect;
}

//==============================================================================
// Set drive, side and density
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void fdc_ext_w(UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (debuglvl > 0)
    fprintf(debuglog, "fdc_ext_w: called\n");
 ctrl_drive = data & 0x03;
 ctrl_side = (data & 0x04) >> 2;
 ctrl_ddense = (data & 0x08) >> 3;
}

//==============================================================================
// Get intrq and drq
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			0x80 if intrq or drq set, else 0
//==============================================================================
UINT16 fdc_ext_r (UINT16 port, struct z80PortRead *port_s)
{
 int status;

 if (ctrl_intrq | ctrl_drq) 
    status = 0x80;
 else
    status = 0x00;
 
 if (debuglvl > 0)
    fprintf(debuglog, "fdc_ext_r: ctrl_intrq, crtl_drq status=%d\n", status);

// printf("fdc_ext_r: ctrl_intrq, crtl_drq status=%d\n", status);

 return status;
}
