//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                                 GUI module                                 *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Provides the graphical interface.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.0.0 - 21 June 2007, S.J.Kay
// - Added fullscreen toggle mode (ALT+ENTER), special code for win32 is used.
// - Added screen resizing capabilities.
//
// v0.0.0 - 5 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//============================================================================== 
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================   

#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>

#include "gui.h"
#include "crtc.h"
#include "ubee512.h"

//==============================================================================
// structures and variables
//==============================================================================
extern int fullscreen;

extern int hdisp;
extern int vdisp;
extern int scans_per_row;

static SDL_Surface *screen;

UINT8 *keystate;

//==============================================================================
// GUI Blit surface.  Display the source surface on the screen surface.
//
//   pass: SDL_Surface src
// return: void
//==============================================================================
void gui_blitsurface (SDL_Surface *src)
{
 SDL_BlitSurface(src, NULL, screen, NULL);
}

//==============================================================================
// GUI Change video.
//
// Will switch to the current mode of window or fullscreen.
//
//   pass: int w		width 
//         int h		height
// return: int			0 if no error, -1 if error
//==============================================================================
int gui_videochange (int w, int h)
{
 if (fullscreen)
    {
     SDL_ShowCursor(SDL_DISABLE);		// don't show the mouse cursor  
     screen = SDL_SetVideoMode(w, h, 16, SDL_FULLSCREEN);
    } 
 else
    {
     SDL_ShowCursor(SDL_ENABLE);		// show the mouse cursor
     screen = SDL_SetVideoMode(w, h, 16, 0);
    } 

 if (screen == NULL)
   {
    fprintf(stderr, "gui_init: SDL_SetVideoMode failed - %s\n",
       SDL_GetError());
    return -1;
   }

 return 0;
}

//==============================================================================
// GUI initialise.
//
// This creates a default video mode of 640 x 275 (80x25) which is suitable for
// CP/M operating systems.  The video mode will be able to be changed later when
// the CRTC registers are changed,  typically this will occur when switching
// between CP/M and Microbee's 64x16 BASIC screen.
//
//   pass: void
// return: int			0 if no error, -1 if error
//==============================================================================
int gui_init (void)
{
 return gui_videochange(VIDWIDTH, VIDHEIGHT);
}

//==============================================================================
// GUI de-initialise.
//
//   pass: void
// return: int			0
//==============================================================================
int gui_deinit (void)
{
 return 0;
}

//==============================================================================
// GUI reset.
//
//   pass: void
// return: int			0
//==============================================================================
int gui_reset (void)
{
 return 0;
}

//==============================================================================
// GUI redraw.
//
//   pass: void
// return: void
//==============================================================================
void gui_redraw (void)
{
 crtc_redraw();
 SDL_UpdateRect(screen, 0, 0, 0, 0);	// all 0s updates the entire screen
}

//==============================================================================
// GUI toggle display mode.
// 
// The SDL_WM_ToggleFullScreen(screen) is only for X11 at present according
// to the documentation,  the win32 build uses standard methods to do this.
//
//   pass: void
// return: int			0 if no error, -1 if error
//==============================================================================
int gui_toggledisplay (void)
{
 fullscreen = (! fullscreen);

#ifdef MINGW
 if (gui_videochange(hdisp * 8, vdisp * scans_per_row) == -1)
    return -1;
 crtc_set_redraw();
 gui_redraw(); 
#else
 if (fullscreen)
    SDL_ShowCursor(SDL_DISABLE);	// don't show the mouse cursor  
 else
    SDL_ShowCursor(SDL_ENABLE);		// show the mouse cursor
 SDL_WM_ToggleFullScreen(screen);
#endif

 return 0;
}

//==============================================================================
// GUI put pixel - (this is inline code)
// 
//   pass: SDL_Surface *surf
//         int x
//         int y
//         Uint32 col
// return: void
//==============================================================================
inline void putpixel (SDL_Surface *surf, int x, int y, Uint32 col)
{
 Uint8 *p = (Uint8 *)surf->pixels + y * surf->pitch + x * 2;
 if ((x < surf->w) && (y < surf->h))
    *(Uint16 *)p = col;
}
