//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                               Keyboard module                              *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Emulates the keyboard by triggering the lpen input on the CRTC when mem_addr
// matches a depressed key.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.2.0 - 22 August 2007, S.J.Kay
// - Added F1 emulator action key to enable data to be read from a tape file.
// - Changed labels PCK_F11, PCK_F12 to PCK_PAGEDOWN and PCK_END
//
// v1.0.0 - 2 August 2007, S.J.Kay
// Combined emulator specific keys into the main key event loop.  Emulator
// fullscreen toggle keys added (ALT+ENTER).
//
// A major rewrite of this module to allow all keys to be emulated correctly
// and to incorporate the Premium cursor keys.  Key repeating at the correct
// speed is now working but this was related to the CRTC vertical blanking
// status code improvement in crtc.c.
//
// The key release code is not ideal at present. It should only release the
// key that was generated with the shift status at the time, so this code will
// incorrectly reset some Microbee key locations and may be an issue for multi
// keypresses (i.e in games).
//
// v0.0.0 - 11 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <SDL.h>

#include "keyb.h"
#include "gui.h"
#include "crtc.h"
#include "vdu.h"
#include "ubee512.h"

#define MB_KEYS	64
#define PC_KEYS	71

// This must be the same as the pc_keys table below it.
enum
   {
    PCK_a,
    PCK_b,
    PCK_c,
    PCK_d,
    PCK_e,
    PCK_f,
    PCK_g,
    PCK_h,
    PCK_i,
    PCK_j,
    PCK_k,
    PCK_l,
    PCK_m,
    PCK_n,
    PCK_o,
    PCK_p,
    PCK_q,
    PCK_r,
    PCK_s,
    PCK_t,
    PCK_u,
    PCK_v,
    PCK_w,
    PCK_x,
    PCK_y,
    PCK_z,
    PCK_LEFTBRACKET,
    PCK_BACKSLASH,
    PCK_RIGHTBRACKET,

    PCK_DELETE,
    
    PCK_1,

    PCK_3,
    PCK_4,
    PCK_5,

    PCK_COMMA,
    
    PCK_PERIOD,
    PCK_SLASH,

    PCK_ESCAPE,
    PCK_BACKSPACE,
    PCK_TAB,
    PCK_KP_ENTER,
    PCK_RETURN,
    PCK_CAPSLOCK,
    PCK_KP_PLUS,
    PCK_SPACE,
#ifdef use_cursorkeys
    PCK_UP,
#else
    PCK_F1,
#endif    
    PCK_LCTRL,
    PCK_RCTRL,
#ifdef use_cursorkeys     
    PCK_DOWN,
    PCK_LEFT,
#else
    PCK_F2,
    PCK_F3,
#endif     
    PCK_F4,
    PCK_F5,
#ifdef use_cursorkeys
    PCK_RIGHT,
#else
    PCK_F6,
#endif
    PCK_LSHIFT,
    PCK_RSHIFT,

    PCK_0,
    PCK_2,
    PCK_6,
    PCK_7,
    PCK_8,
    PCK_9,
    PCK_BACKQUOTE,
    PCK_QUOTE,
    PCK_SEMICOLON,
    PCK_MINUS,
    PCK_EQUALS,

    PCK_PAGEDOWN,
    PCK_END,
    PCK_LALT,
    PCK_RALT,
    PCK_F1
   };

// PC Keys to be checked for events
SDLKey pc_keys [PC_KEYS] =
   {
    SDLK_a,             // a      01
    SDLK_b,             // b      02
    SDLK_c,             // c      03
    SDLK_d,             // d      04
    SDLK_e,             // e      05
    SDLK_f,             // f      06
    SDLK_g,             // g      07
    SDLK_h,             // h      08
    SDLK_i,             // i      09
    SDLK_j,             // j      0A
    SDLK_k,             // k      0B
    SDLK_l,             // l      0C
    SDLK_m,             // m      0D
    SDLK_n,             // n      0E
    SDLK_o,             // o      0F
    SDLK_p,             // p      10
    SDLK_q,             // q      11
    SDLK_r,             // r      12
    SDLK_s,             // s      13
    SDLK_t,             // t      14
    SDLK_u,             // u      15
    SDLK_v,             // v      16
    SDLK_w,             // w      17
    SDLK_x,             // x      18
    SDLK_y,             // y      19
    SDLK_z,             // z      1A
    SDLK_LEFTBRACKET,   // [{     1B 
    SDLK_BACKSLASH,     // \|     1C
    SDLK_RIGHTBRACKET,  // ]}     1D

    SDLK_DELETE,        // DEL    1F    
    
    SDLK_1,             // 1!     21

    SDLK_3,             // 3#     23
    SDLK_4,             // 4$     24
    SDLK_5,             // 5%     25        

    SDLK_COMMA,         // ,<     2C
    
    SDLK_PERIOD,        // .>     2E
    SDLK_SLASH,         // /?     2F            

    SDLK_ESCAPE,        // ESC    30
    SDLK_BACKSPACE,     // BS     31
    SDLK_TAB,           // TAB    32
    SDLK_KP_ENTER,      // KP_ENT 33
    SDLK_RETURN,        // ENT    34  
    SDLK_CAPSLOCK,      // CAPL   35 
    SDLK_KP_PLUS,       // KP_+   36 
    SDLK_SPACE,         // SP     37 
#ifdef use_cursorkeys
    SDLK_UP,		// UP     38
#else
    SDLK_F1,            // F1     38
#endif
    SDLK_LCTRL,         // LCTRL  39 
    SDLK_RCTRL,         // RCTRL  39      
#ifdef use_cursorkeys     
    SDLK_DOWN,		// DOWN   3A
    SDLK_LEFT,		// LEFT   3B
#else
    SDLK_F2,            // F2     3A 
    SDLK_F3,            // F3     3B
#endif        
    SDLK_F4,            // F4     3C    
    SDLK_F5,            // F5     3D 
#ifdef use_cursorkeys
    SDLK_RIGHT,		// RIGHT  3E    
#else
    SDLK_F6,            // F6     3E 
#endif
    SDLK_LSHIFT,        // LSHIFT 3F 
    SDLK_RSHIFT,        // RSHIFT 3F  

// The following keys change for upper and lower case
//                                L   INV    U    INV
    SDLK_0,             // 0)   0-20   -    )-29   -
    SDLK_2,             // 2@   2-22   -    @-00   Y
    SDLK_6,             // 6^   6-26   -    ^-1E   Y
    SDLK_7,             // 7&   7-27   -    &-26   -
    SDLK_8,             // 8*   8-28   -    *-2A   -
    SDLK_9,             // 9(   9-29   -    (-28   -
    SDLK_BACKQUOTE,     // `~   `-00   Y    ~-1E   -
    SDLK_QUOTE,         // '"   '-27   Y    "-22   -
    SDLK_SEMICOLON,     // ;:   ;-2B   -    :-2A   Y
    SDLK_MINUS,         // -_   --2D   -    _-1F   -
    SDLK_EQUALS,        // =+   =-2D   Y    +-2B   -

// The following keys are only for emulator usage
    SDLK_PAGEDOWN,      // PAGEDOWN n/a
    SDLK_END,           // END      n/a
    SDLK_LALT,          // LALT     n/a
    SDLK_RALT,          // RALT     n/a
    SDLK_F1             // F1       n/a
   };

// Microbee codes for lower PC keys.  If a key is not implemented for
// the Microbee key use 0xff.
UINT8 mb_scan_pclower [PC_KEYS] =
{
 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x10,
 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18,
 0x19, 0x1A, 0x1B, 0x1C, 0x1D,
 0x1F,
 0x21,
 0x23, 0x24, 0x25,
 0x2C,
 0x2E, 0x2F,
 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
 0x38, 0x39, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E,
 0x3F, 0x3F,
 0x20, 0x22, 0x26, 0x27, 0x28, 0x29, 0x00, 0x27,
 0x2B, 0x2D, 0x2D,
 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

// Microbee codes for shift PC keys.  If a key is not implemented for
// the Microbee key use 0xff.
UINT8 mb_scan_pcshift [PC_KEYS] =
{
 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08,
 0x09, 0x0A, 0x0B, 0x0C, 0x0D, 0x0E, 0x0F, 0x10,
 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17, 0x18,
 0x19, 0x1A, 0x1B, 0x1C, 0x1D,
 0x1F,
 0x21,
 0x23, 0x24, 0x25,
 0x2C,
 0x2E, 0x2F,
 0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,
 0x38, 0x39, 0x39, 0x3A, 0x3B, 0x3C, 0x3D, 0x3E,
 0x3F, 0x3F,
 0x29, 0x00, 0x1E, 0x26, 0x2A, 0x28, 0x1E, 0x22,
 0x2A, 0x1F, 0x2B,
 0xFF, 0xFF, 0xFF, 0xFF, 0xFF
};

// PC status of all keys required for emulation. 0 if key up or 1 if key down.
static UINT8 pc_keystate[PC_KEYS];

// MB status of all keys required for emulation. 0 if key up or 1 if key down.
static UINT8 mb_keystate[MB_KEYS];

// PC shift used for all Microbee keys
static UINT8 mb_invert[MB_KEYS];

extern int key_quit;
extern int key_reset;
extern int key_tape;

static SDL_Event event;

static int stopshift;
static int makeshift;

static int savescan;
static int forcescans;
static int havekeys;

static char capslock_a;

extern FILE *debuglog;
extern int debuglvl;

//==============================================================================
// Keyboard initialise
//
//   pass: void
// return: int 				0
//==============================================================================
int keyb_init (void)
{
 return 0;
}

//==============================================================================
// Keyboard de-initialise
//
//   pass: void
// return: int 				0
//==============================================================================
int keyb_deinit (void)
{
 return 0;
}

//==============================================================================
// Keyboard reset
//
//   pass: void
// return: int 				0
//==============================================================================
int keyb_reset (void)
{
 return 0;
}

//==============================================================================
// Get the down status of a Microbee key.
//
//   pass: int scan			scan code 0-63
// return: int getkeystate		1=key is down, else 0
//==============================================================================
int getkeystate (int scan)
{
 static int last_capslock = 0;
 static int capslock_count = 3;

 int shift;

#if 1
 if ((scan == 0x35) && (! capslock_a)) 		// if CAPSLOCK
    {
     if (mb_keystate[0x35] != last_capslock)
        {
         if (--capslock_count == 0)
            {
             last_capslock = mb_keystate[0x35];
             capslock_count = 3;
            }
         return 1;
        }
     return 0;
    }
#endif    

// if force scans then return state as 1 for required counts 
 if (forcescans)
    {
     if (savescan == scan)
        {
         if (--forcescans);
         return 1;
        }
     return 0;
    }    

// return 0 if stopping shift key actions
 if ((stopshift) && (scan == MB_KEYS-1))
    {
     stopshift--;        
     return 0;
    } 

// return 1 if making shift key actions        
 if ((makeshift) && (scan == MB_KEYS-1))
    {
     makeshift--;
     return 1;
    } 

 shift = mb_keystate[0x3F];

 if (mb_keystate[scan])
    {
     if (shift)
        {
#ifdef use_debug_getkeystate_shift
         printf("SHIFT - mb scan=%d  invert=%d\n", scan, mb_invert[scan]);
#endif
         if (! mb_invert[scan])
            return 1;             	// return a normal key
         else
            {
             stopshift = 25;      	// wer'e inverting to lower !
             return 1;
            }
        }   
     else
        {
#ifdef use_debug_getkeystate_lower
         printf("LOWER - mb scan=%d  invert=%d\n", scan, mb_invert[scan]);        
#endif
         if (! mb_invert[scan])
            return 1;             	// return a normal key
         else
            {
             makeshift = 25;       	// wer'e inverting to SHIFT !
             return 1;
            }
        }
    }    
 else
    return 0;
}

//==============================================================================
// Check for any new key events and serive all that are found.  This creates
// the Microbee scan keys and determines what inverting action may be required
// for each Microbee key generated.
//
//   pass: void
// return: void
//==============================================================================
void keyb_keysupdate (void)
{
 SDLKey key;
 int i; 
 int scan;
 int shift;
 
// Poll for events. SDL_PollEvent() returns 0 when there are no
// more events on the event queue.
 
 while (SDL_PollEvent(&event))
    {
     key = event.key.keysym.sym;    

     for (i = 0; i < PC_KEYS; i++)
        {
         if (key == pc_keys[i])
            {
             switch (event.type)
                {
                 case SDL_KEYDOWN:
                      pc_keystate[i] = 1;
                      shift = (pc_keystate[PCK_LSHIFT] | pc_keystate[PCK_RSHIFT]);
                      if (shift)
                         {
                          scan = mb_scan_pcshift[i];
                          if (scan != 255)
                             {
                              havekeys++; 
                              if ((key == SDLK_2) || (key == SDLK_6) || (key == SDLK_SEMICOLON))
                                 mb_invert[scan] = 1;
                              else
                                 mb_invert[scan] = 0;
                             }    
                         } 
                      else
                         {
                          scan = mb_scan_pclower[i];
                          if ((key == SDLK_RETURN) && (pc_keystate[PCK_LALT] || pc_keystate[PCK_RALT]))
                              {
                               gui_toggledisplay();
                               break;
                              } 
                          if (scan != 255)
                             {
                              havekeys++; 
                              if ((key == SDLK_BACKQUOTE) || (key == SDLK_QUOTE) || (key == SDLK_EQUALS))
                                 mb_invert[scan] = 1;
                              else
                                 mb_invert[scan] = 0;
                             }
                          else		
                             {
                              switch (i)	// Emulator specific operation keys
                                 {
                                  case  PCK_PAGEDOWN : key_reset = 1;
                                                       break;
                                  case       PCK_END : key_quit = 1;
                                                       break;
                                  case        PCK_F1 : key_tape = 2;
                                                       break;                                                                                                        
                                 }
                             }   
                         } 
                      if (scan != 255)
                         mb_keystate[scan] = 1;
                      break;
                 case SDL_KEYUP:
                      pc_keystate[i] = 0;                 
                      if ((scan != 255) && (havekeys))
                         havekeys--;
// shift = (pc_keystate[PCK_LSHIFT] | pc_keystate[PCK_RSHIFT]);
// This following is not ideal as some multikey presses may be stopped.
                      if (mb_scan_pcshift[i] != 255)
                         mb_keystate[mb_scan_pcshift[i]] = 0;
                      if (mb_scan_pclower[i] != 255)
                         mb_keystate[mb_scan_pclower[i]] = 0;
                      break;
                }
            }       
        }
    }
}

//==============================================================================
// Scan 1 key at the address past.  If key is detected then the light pen
// register strobe is set.  This is called when the crtc lpen address has been
// set.
//
//   pass: int addr			address of light pen
// return: void
//==============================================================================
void keyb_handler (int addr)
{
 int scan;

 scan = (addr >> 4) & 0x3F;

 if (debuglvl > 0)
    fprintf(debuglog, "keyb: addr = 0x%04x, scan = 0x%02x\n", addr, scan);

 keyb_keysupdate();

// if (havekeys) 
    if (getkeystate(scan))
       crtc_lpen(addr);
}

//==============================================================================
// Scan all keys and setting the light pen registers to first found depressed
// key. This is called from the crtc.c for the 6545 status register.
//
//   pass: void
// return: void
//==============================================================================
void keyb_checkall (void)
{
 int i;

 keyb_keysupdate();

 if (! latchrom) 
//  if ((! latchrom) && (havekeys))
    {
     for (i = 0; i < MB_KEYS; i++)
        if (getkeystate(i))
           {
            crtc_lpen(i << 4);
            break;
           }
    }
}
