//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                            Memory mapper module                            *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Emulates the 512K memory bank switching and video active signal.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.0.0 - 7 July 2007, S.J.Kay
// - added vdu_switchvideo_in() and vdu_switchvideo_out() function calls.
// - ROMs 1-3 required seperate handlers as there will be DRAM at 0xC000-0xDFFF
//   when ROM3 is selected.  This change allows the DRAM area to be written to.
//   A smaller dupwrite handler is also activated if the top and bottom banks
//   are the same and ROM3 is selected. (there is an 8K DRAM window).
// - According to the Technical manual ROM3 appears to be placed into wrong
//   memory location (0xC000) and should be placed into 0xE000.  Also ROM3 is
//   documented as being 8K and ROM1 and ROM2 are 16K.  Have made changes in
//   various locations for this and have modified the load_rom function and
//   optioned out the roundup function for now.
//
// - Converted to 512K DRAM model Microbee (as used for PJB 512K BIOS).  Now
//   uses array of pointers for banks.
//
// v0.0.0 - 5 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "memmap.h"
#include "ubee512.h"
#include "z80.h"
#include "vdu.h"
#include "mz80.h"

#include "macros.h"

static int loadrom(char *name, char *dest, int size);     /* load a rom image */

static void switchin(int newmode);
static void switchout(int oldmode);

static void rom1write(UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);
static void rom2write(UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);
static void rom3write(UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);
static void dupwrite(UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s);

static struct MemoryWriteByte z80_mem_w[MAXMEMHANDLERS] =
   { { -1, -1, NULL, NULL } };

static struct MemoryReadByte z80_mem_r[MAXMEMHANDLERS] =
   { { -1, -1, NULL, NULL } };

static char rom1[ROM1_SIZE],
            rom2[ROM2_SIZE],
            rom3[ROM3_SIZE];

static char bank00[BANKSIZE],
            bank01[BANKSIZE],
            bank02[BANKSIZE],
            bank03[BANKSIZE],
            bank04[BANKSIZE],
            bank05[BANKSIZE],
            bank06[BANKSIZE],
            bank07[BANKSIZE],
            bank08[BANKSIZE],
            bank09[BANKSIZE],
            bank10[BANKSIZE],
            bank11[BANKSIZE],
            bank12[BANKSIZE],
            bank13[BANKSIZE],
            bank14[BANKSIZE],
            bank15[BANKSIZE];

char const *bank_ptrs[] =
   {
    bank01, bank02, bank00, bank03,
    bank00, bank02, bank01, bank03,

    bank04, bank05, bank06, bank07,
    bank04, bank05, bank06, bank07,
    bank08, bank09, bank10, bank11,
    bank08, bank09, bank10, bank11,
    bank12, bank13, bank14, bank15,
    bank12, bank13, bank14, bank15
   };

char const *dram_bank_ptrs[] =
   {
    bank01, bank02, bank00, bank03,
    bank04, bank05, bank06, bank07,
    bank08, bank09, bank10, bank11,
    bank12, bank13, bank14, bank15
   };

int mem_mode;		// current memory map configuration (port 0x50)

extern char romimage1[];
extern char romimage2[];
extern char romimage3[];

extern FILE *debuglog;
extern int debuglvl;

#if 0
//==============================================================================
// rounds up to a power of 2
//
//   pass: int n
// return: unsigned int			rounded up int value
//==============================================================================
static unsigned int roundup(unsigned int n)
{
 int bits_set = 0;
 int high_bit = 0;
 int i;
 int n2 = n;

 for (i=0; i<(sizeof(unsigned int)*8); i++)
    {
     if (n2 & 1)
        {
         high_bit = i;
         bits_set++;
        }
     n2 >>= 1;
    }

 if (bits_set > 1)
    return 1 << (high_bit+1);
 else
    return n;
}
#endif

//==============================================================================
// Load a ROM image from file
//
//   pass: char *name  		pointer to file name
//         char *dest  		pointer to destination
//         int size 		expected size of the ROM to be loaded
// return: int			0 if no errors, -1 if error
//==============================================================================
static int loadrom(char *name, char *dest, int size)
{
 FILE *romfp;
 long len;

 if (dest == NULL)
    return -1;

 memset(dest, 0, size);

 romfp = fopen(name, "rb");
 if (romfp == NULL)
    return -1;

 fseek(romfp, 0, SEEK_END);  		// get size of ROM
 len = ftell(romfp);
 fseek(romfp, 0, SEEK_SET);		// back to start
 
 if (len > size)
    len = size;

 if (len == 0)
    return -1;

 fread(dest, 1, len, romfp);
 fclose(romfp);

 return 0;
}

//==============================================================================
// Initialise the memory map
//
//   pass: void
// return: int			0 if no errors, -1 if error
//==============================================================================
int memmap_init(void)
{
 struct mz80context z80;

// For debugging purposes only, fills each DRAM bank with numbers.
#ifdef use_dramdebug
 int i;
 for (i = 0; i < 16; i++)
    memset(dram_bank_ptrs[i], i, BANKSIZE);
#endif

 if (loadrom(romimage1, rom1, ROM1_SIZE) != 0)
    {
     fprintf(stderr, "Unable to load rom1.bin (required to boot).\n");
     return -1;
    }
 loadrom(romimage2, rom2, ROM2_SIZE);
 loadrom(romimage3, rom3, ROM3_SIZE);   
    
 mz80GetContext(&z80);
 z80.z80MemRead = z80_mem_r;
 z80.z80MemWrite = z80_mem_w;
 mz80SetContext(&z80);

// initialise the mamory map to default
 mem_mode = -1; 

 memmap_mode(0x50, 0, NULL);

 return 0;
}

//==============================================================================
// De-init the memory map (N/A)
//
//   pass: void
// return: int			0
//==============================================================================
int memmap_deinit(void)
{
 return 0;
}

//==============================================================================
// Set memory map to the reset condition
//
//   pass: void
// return: int			0
//==============================================================================
int memmap_reset(void)
{
 memmap_mode(0x50, 0, NULL);
 return 0;
}

//==============================================================================
// Set memory map mode - Port 0x50 write handler
//
//   pass: UINT16 port
//	   UINT8 data
//         struct z80PortWrite *port_struct
// return: void
//==============================================================================
void memmap_mode(UINT16 port, UINT8 data, struct z80PortWrite *port_struct)
{
 if (debuglvl > 0)
    fprintf(debuglog, "memmap_mode set = 0x%02x\n", data);

#ifdef use_debug_memmap_mode
   printf("memmap_mode set = 0x%02x\n", data);
#endif

 if (mem_mode != -1)
    switchout(mem_mode);

#ifdef use_dram512
// SJK 5 June 07 - more bits needed for 512K (11xxxx11) 
// switchin(mem_mode = (data & B8(11011111)));
 switchin(mem_mode = (data & B8(11111111))); 
#else
// switchin(mem_mode = (data & B8(00011111)));
 switchin(mem_mode = (data & B8(00111111))); 
#endif
}

//==============================================================================
// Switch in DRAM banks, Video memory locations, ROMs 1,2 and 3, etc
//
//   pass: int newmode
// return: void
//==============================================================================
static void switchin (int newmode)
{
 int rindex = 0;
 int windex = 0;
 int banksel;
 char *bank;

#ifdef use_dram512
 banksel = ((newmode & B8(11000000)) >> 3) | (newmode & B8(00000111));
#else
 newmode &= B8(00111111);
 banksel = newmode & B8(00000111);
#endif

 bank = (char *)bank_ptrs[banksel];

// Copy DRAM bank to memory at 0x0000-0x7FFF
 memcpy(z80mem, bank, BANKSIZE);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// insert video memory handler if video RAM is enabled.
//
// Must appear early in the list of handlers so it overrides anything that
// overlaps it.  The handler that gets called is the first one with matching
// range, as a result of the mz80 implementation.  This might change in the
// future!
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 if (! (newmode & BANK_VRAM))	// if video RAM enabled
   {
    z80_mem_r[rindex].memoryCall = vdu_vidmem_r;
    z80_mem_r[rindex].pUserArea  = NULL;

    z80_mem_w[windex].memoryCall = vdu_vidmem_w;
    z80_mem_w[windex].pUserArea  = NULL;

    if (newmode & BANK_VADD)   // if video RAM at 8000-8FFF
       {
        z80_mem_r[rindex].lowAddr =  0x8000;
        z80_mem_r[rindex].highAddr = 0x8FFF;        

        z80_mem_w[windex].lowAddr =  0x8000;
        z80_mem_w[windex].highAddr = 0x8FFF;
       }
    else
       {
        z80_mem_r[rindex].lowAddr =  0xF000;
        z80_mem_r[rindex].highAddr = 0xFFFF;

        z80_mem_w[windex].lowAddr =  0xF000;
        z80_mem_w[windex].highAddr = 0xFFFF;
       }
    rindex++;
    windex++;
   }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Test if any ROMs are in and set up handlers accordingly.
//
// - if no ROMs in then copy the complete memory bank BANKUBLK to upper
//   memmory.
// - if ROM3 is in then there is an 8K window of DRAM at 0xC000-DFFF and this
//   copyed to from BANKUBLK.
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 if (newmode & BANK_ROMS)     		// if ROMs disabled
    {
     memcpy(z80mem+BANKSIZE, BANKUBLK, BANKSIZE);  // copy bank

     if ((newmode & BANK_DRAM) == BANK_HIGH)  // if high bank
        {
// this bank is mapped into upper and lower memory , so we need to echo any
// writes to one half to the other.
         z80_mem_w[windex].lowAddr = 0x0000;
         z80_mem_w[windex].highAddr = 0xFFFF;
         z80_mem_w[windex].memoryCall = dupwrite;
         z80_mem_w[windex].pUserArea = NULL;
         windex++;
        }
    }
 else
    {
     z80_mem_w[windex].lowAddr = 0x8000;
     z80_mem_w[windex].highAddr = 0xBFFF;
     z80_mem_w[windex].memoryCall = rom1write;
     z80_mem_w[windex].pUserArea = NULL;
     windex++;    
     memcpy(z80mem+0x8000, rom1, ROM1_SIZE);
     
     if (newmode & BANK_ROMX)   // if ROM3 selected
        {
         memcpy(z80mem+0xC000, BANKUBLK+0x4000, 0x2000);  // copy 8K only        

         if ((newmode & BANK_DRAM) == BANK_HIGH)  // if high bank
            {
// this bank is mapped into upper and lower memory , so we need to echo any
// writes to one half to the other.
             z80_mem_w[windex].lowAddr = 0x04000;
             z80_mem_w[windex].highAddr = 0x5FFF;
             z80_mem_w[windex].memoryCall = dupwrite;
             z80_mem_w[windex].pUserArea = NULL;
             windex++;
            }
         z80_mem_w[windex].lowAddr = 0xC000; // need to handle some DRAM too
         z80_mem_w[windex].highAddr = 0xFFFF;
         z80_mem_w[windex].memoryCall = rom3write;
         z80_mem_w[windex].pUserArea = NULL;
         windex++;     
         memcpy(z80mem+0xE000, rom3, ROM3_SIZE);
        } 
     else
        {
         z80_mem_w[windex].lowAddr = 0xC000;
         z80_mem_w[windex].highAddr = 0xFFFF;
         z80_mem_w[windex].memoryCall = rom2write;
         z80_mem_w[windex].pUserArea = NULL;
         windex++;        
         memcpy(z80mem+0xC000, rom2, ROM2_SIZE);
        } 
    }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Video memory.
//
// Copy video memory into the main memory if video is enabled, done last so it
// overwrites anything already there.
//
// NOTE: use of vb & B8(00000011) for screen, attribute and colour RAM as
// these are at most 8K each (4 x 2K banks).
//
// The vb value used here will be used in the switch out function so that the
// correct video RAM banks can be switched.
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 if (! (newmode & BANK_VRAM))	// if video RAM enabled
    {
#ifdef use_alpha_p
     vdu_switchvideo_in();
#else
     if (newmode & BANK_VADD)  	// if video at 8000-8FFF
        {
         memcpy(z80mem+0x8000, videoram, 0x800);
         memcpy(z80mem+0x8800, pcgram, 0x800);
        }
     else
        {
         memcpy(z80mem+0xF000, videoram, 0x800);
         memcpy(z80mem+0xF800, pcgram, 0x800);
        }
#endif
    }

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// terminate the memory handlers
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 z80_mem_r[rindex].lowAddr = -1;
 z80_mem_r[rindex].highAddr = -1;
 z80_mem_r[rindex].memoryCall = NULL;
 z80_mem_r[rindex].pUserArea = NULL;

 z80_mem_w[windex].lowAddr = -1;
 z80_mem_w[windex].highAddr = -1;
 z80_mem_w[windex].memoryCall = NULL;
 z80_mem_w[windex].pUserArea = NULL;
}

//==============================================================================
// Switch out DRAM banks, Video memory locations, ROMs 1,2 and 3, etc
//
//   pass: int oldmode
// return: void
//==============================================================================
static void switchout(int oldmode)
{
 int banksel;
 char *bank;

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Copy DRAM memory at 0000-7FFF to bank
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
#ifdef use_dram512
 banksel = ((oldmode & B8(11000000)) >> 3) | (oldmode & B8(00000111));
#else
 oldmode &= B8(00111111);  //SJK
 banksel = oldmode & B8(00000111); 
#endif        

 bank = (char *)bank_ptrs[banksel];
 
// Copy DRAM memory at 0000-7FFF to bank
 memcpy(bank, z80mem, BANKSIZE);

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// Switch video RAM out if enabled.
//
// NOTE: use of vb & B8(00000011) for screen, attribute and colour RAM as
// these are at most 8K each (4 x 2K banks).
//
// The vb value used in the switch in function will be used in the switch out
// so that the correct video RAM banks can be switched.
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 if (!(oldmode & BANK_VRAM))  // if video RAM enabled
    {
#ifdef use_alpha_p
     vdu_switchvideo_out();

     if (oldmode & BANK_VADD) // if video at 8000-8FFF
        {
         if (oldmode & BANK_ROMS)	// if ROMs disabled
            {
             if ((oldmode & BANK_DRAM) != BANK_HIGH)  // if NOT high bank
                memcpy(BANKUBLK+0x1000, z80mem+0x9000, 0x7000);
            }
         else
            {
             if (((oldmode & BANK_DRAM) != BANK_HIGH) && (oldmode & BANK_ROMX))   // if ROM3 selected             
                memcpy(BANKUBLK+0x4000, z80mem+0xC000, 0x2000);                   
            }    
        }
     else
        {
         if (oldmode & BANK_ROMS)    	// if ROMs disabled
            {
             if ((oldmode & BANK_DRAM) != BANK_HIGH)  // if NOT high bank
                memcpy(BANKUBLK, z80mem+0x8000, 0x7000);
            }
         else
            { 
             if (((oldmode & BANK_DRAM) != BANK_HIGH) && (oldmode & BANK_ROMX))   // if ROM3 selected             
                memcpy(BANKUBLK+0x4000, z80mem+0xC000, 0x2000);                   
            }       
        }
#else
     if (oldmode & BANK_VADD) // if video at 8000-8FFF
        {
         memcpy(videoram, z80mem+0x8000, 0x800);
         memcpy(pcgram, z80mem+0x8800, 0x800);

         if (oldmode & BANK_ROMS)	// if ROMs disabled
            if ((oldmode & BANK_DRAM) != BANK_HIGH)  // if NOT high bank
               memcpy(BANKUBLK+0x1000, z80mem+BANKSIZE+0x1000, BANKSIZE-0x1000);
        }
     else
        {
         memcpy(videoram, z80mem+0xF000, 0x800);
         memcpy(pcgram, z80mem+0xF800, 0x800);

         if (oldmode & BANK_ROMS)    	// if ROMs disabled
            if ((oldmode & BANK_DRAM) != BANK_HIGH)  // if NOT high bank
               memcpy(BANKUBLK, z80mem+BANKSIZE, BANKSIZE-0x1000);
        }
#endif
    }
 else

//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
// no video RAM, or ROMS enabled, copy current memory map 8000-FFFF back to it's
// bank location.
//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    if (oldmode & BANK_ROMS) 		// if ROMs are disabled
       if ((oldmode & BANK_DRAM) != BANK_HIGH)  // if NOT high bank
// copy memory map 8000-FFFF to bank
          memcpy(BANKUBLK, z80mem+BANKSIZE, BANKSIZE);         
}

//==============================================================================
// ROM 1 write (0x8000-0xBFFF)
//
// This handler is installed when ROM1 (port 0x50 bit 2=0) are in the memory
// map. No writing to any ROM areas will occur.
//
//   pass: UINT32 addr
//	   UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void rom1write(UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
}

//==============================================================================
// ROM 2 write (0xC000-0FFFF)
//
// This handler is installed when ROM2 (port 0x50 bit 2=0) and the BANK_ROMX
// bit is clear in mem_mode.  No writing to any ROM areas will occur.
//
//   pass: UINT32 addr
//	   UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void rom2write(UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
}

//==============================================================================
// ROM 3 write (0xE000-0FFFF)
//
// This handler is installed when ROM3 (port 0x50 bit 2=0) and the BANK_ROMX
// bit is set in mem_mode.  No writing to any ROM areas will occur.
//
// If the memory access address is DRAM (0xC000-0xDFFF) then a write will
// occur to DRAM memory.
//
// If BANKUBLK is in context at 0x0000 then any DRAM write (0xC000-0xDFFF)
// must also be written to DRAM (0x4000-0x5FFF)
//
//   pass: UINT32 addr
//	   UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void rom3write(UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
 if ((addr > 0xBFFF) && (addr < 0xE000))
    {
     z80mem[addr] = data;
     if ((mem_mode & BANK_DRAM) == BANK_HIGH)  // if high bank
        z80mem[addr & 0x7FFF] = data;
    }
}

//==============================================================================
// duplicates writes to one half of DRAM to the other half.  This handler is
// installed when DRAM bank 0, block 1 is in the lower half of memory as it is
// also in the upper half of memory.  Depending on the selection of ROM3
// the address range for this handler activation is different.
//
//   pass: UINT32 addr
//         UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
static void dupwrite(UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
 z80mem[addr & 0x7FFF] = data;
 z80mem[addr | 0x8000] = data;
}
