//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                              Z80 PIO module                                *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Emulate the Z80 PIO functionality and interrupts.
//
// All PIO peripheral devices are maintained in seperate source files:
//
// printer.c		Parallel printer port
// tape.c		Tape in/out port
// serial.c		Serial in/out port
// sound.c		Speaker sound
// rtc.c		RTC (interrupt)
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.3.0 - 1 September 2007, S.J.Kay
// - Emulate PIO internal registers and interrupts.
// - Added RS232 serial TX, RX, DTR and CTS.
// - Added printer (parallel) out on port A.
//
// v1.2.0 - 22 August 2007, S.J.Kay
// - Added tape out and tape in.
//
// v1.0.0 - 23 June 2007, S.J.Kay
// - Created a new file to implement the Z80 PIO emulation.

//==============================================================================
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <inttypes.h>
#include <string.h>
#include <time.h>
#include <SDL.h>

#include "pio.h"
#include "sound.h"
#include "tape.h"
#include "serial.h"
#include "printer.h"
#include "ubee512.h"
#include "mz80.h"

#include "macros.h"

pio_t pio_a;
pio_t pio_b;

static int execratio_porta;
static int execratio_serial;

extern int z80_ratio;
 
extern char key_tape;

extern int serial_byte_rx;

extern char *tape_i;
extern char *tape_o;

extern char *print_a;
extern char *print_b;

extern char *coms1_dev;
extern int coms1;

extern FILE *debuglog;
extern int debuglvl;

//==============================================================================
// PIO - Control register
//
//   pass: pio_t *pio
//         UINT8 data
//         char port
// return: void
//==============================================================================
static void pio_control (pio_t *pio, UINT8 data, char port)
{
 pio->cont = data;
                 
 if (pio->action)
    {
     if (pio->action == 1)
        {     
         pio->direction = data; // set data direction (1=input, 0=output)
#ifdef use_pioinfo
         printf("Port %c: set data direction: 0x%02x\n", port, data);
#endif
        }
     if (pio->action == 2)
        {
         pio->maskword = data;
#ifdef use_pioinfo
         printf("Port %c: mask word: 0x%02x\n", port, data);
#endif
        }
     pio->action = 0;   
    }
 else
    {
     if ((data & B8(00000001)) == 0) // if interrupt vector
       {
        pio->vector = data;
#ifdef use_pioinfo
        printf("Port %c: interrupt vector: 0x%02x\n", port, data);
#endif        
       } 
     else   
        if ((data & B8(00001111)) == B8(00000111)) // if interrupt control word
           {
            pio->ienable = data & B8(10000000);
            if (pio->ienable)
               pio->ienableff = B8(10000000);               
            pio->andor = data & B8(01000000);
            pio->hilo = data & B8(00100000);
            if (data & B8(00010000)) // if mask word follows
               {
                pio->action = 2;
                pio->pending = 0;
               }
#ifdef use_pioinfo
            printf("Port %c: interrupt control word: 0x%02x\n", port, data);
#endif            
           }    
        else
           if ((data & B8(00001111)) == B8(00001111)) // if mode select
              {
               pio->mode = (data & B8(11000000)) >> 6;
               if (pio->mode == 3) // if mode 3 (bit control)
                  pio->action = 1;
#ifdef use_pioinfo
               printf("Port %c: mode select: 0x%02x\n", port, data);
#endif               
              }
           else   
              if ((data & B8(00001111)) == B8(00000011)) // if interrupt flipflop set
                 {
                  pio->ienableff = data & B8(10000000);
#ifdef use_pioinfo
                  printf("Port %c: interrupt flipflop set: 0x%02x\n", port, data);
#endif
                 }
    }    
}

//==============================================================================
// PIO Initialise
//
//   pass: void
// return: int				0 if success, -1 if error
//==============================================================================
int pio_init (void)
{
 int x=0;
#ifdef use_audio
 x = sound_init();
#endif
 return (x | tape_init() | serial_init() | printer_init());
}

//==============================================================================
// PIO de-initialise
//
//   pass: void
// return: int				0 if success, -1 if error
//==============================================================================
int pio_deinit (void)
{
 int x=0;
#ifdef use_audio
 x = sound_deinit();
#endif
 return (x | tape_deinit() | serial_deinit() | printer_deinit()); 
}

//==============================================================================
// PIO reset
//
//   pass: void
// return: int				0 if success, -1 if error
//==============================================================================
int pio_reset (void)
{
 int x=0;
 
 pio_a.mode = 1;
 pio_a.maskword = 0; 
 pio_a.ienableff = 0;
 pio_a.data = 0; 
 pio_a.action = 0; 

 pio_b.mode = 1;
 pio_b.maskword = 0; 
 pio_b.ienableff = 0;
 pio_b.data = 0; 
 pio_b.action = 0; 
 
#ifdef use_audio
 x = sound_reset();
#endif
 return (x | tape_reset() | serial_reset() | printer_reset());
}

//==============================================================================
// PIO - poll for interrupt events.
//
// Checks to see if a maskable interrupt should be generated,  if so, set the
// interrupt to occur for the next mz80exec() block.
//
// A special z80_exec_ratio() function is used to allow faster responses to
// multiple interrupts.
//
// The interrupts are limited at present as the full method to determine
// what causes an interrupt is still to be implemented correctly but what is
// here should work for standard things, no clock interrupt on port B bit 7
// or tape at this stage.
//
// Interrupts supported:
//
// RS232 Serial interrupts
// -----------------------
// If a byte has been received and ready in a buffer and the last RX PIO bit
// was a stop bit (serial_byte_rx == -1 in our case) an interrupt will be
// generated.
//
// See the pio_configure() function for information about the exec_ratio
// value used here.
//
// Port A LPT
// ----------
// Generates an interrupt after each write to port A.
//
//   pass: void
// return: void
//==============================================================================
void pio_polling (void)
{
// if (pio_a.ienable)
 if (pio_a.ienableff)
    {
     if (printer_strobe_get() == 0)
        {
         printer_strobe_set();
         z80_exec_ratio(execratio_porta);
         mz80int(pio_a.vector);
#ifdef use_pioinfo
         printf("pio_interrupt (A): vector=0x%02x\n", pio_a.vector);
#endif
         return;
        }
    }     

 if (pio_b.ienableff)
    {
     if (serial_byte_rx == -1)
        {
         if (serial_readpoll() != -1)
            {
             z80_exec_ratio(execratio_serial);
             mz80int(pio_b.vector);
#ifdef use_pioinfo
             printf("pio_interrupt (B): vector=0x%02x\n", pio_b.vector);
#endif
             return;
            }
        }
     else
        serial_r();	// see explanation in this function why we call
    } 

// set the Z80 execution ratio back to normal,  i.e 30 interrupts/sec @
// 3.375 MHz if no interrupt events.
 if (z80_ratio != 1)
    z80_exec_ratio(1);
 return;
} 

//==============================================================================
// PIO - configure.
//
// RS232 Serial interrupts
// -----------------------
// A z80_exec_ratio(40) ratio was found to be about right at fast baud rates
// values of 60 did not work any faster, 70 did not work in some circumstances.
//
// To get a true through put rate of 19200 baud (1920 bytes/sec) would require
// 1920 interrupts per second.  Some more experimentation is needed with
// the ratio value but for now a value of 40 is being used as maximum.
//
// At slower baud rates having too high a ratio will slow response time down
// so the ratio needs to be determined by the RX baud rate:
//
// These calculations are based on default values for the frame rate and XTAL:
//
// the time for 1 Z80 frame is: tz = (3375000 / 30) / 3375000
// the time for 1 serial bit is: tb = (1 / baudrx)
//
// ratio = tz / tb
//
//   pass: int xtal
//         int fram_rate
//         int serial_rx_baud
// return: void
//==============================================================================
void pio_configure (int xtal, int frame_rate, int serial_rx_baud)
{
 float tz;
 float tb;

// set port A to about 600 interrupts/sec @ 3.375 MHz
 execratio_porta = 20;

// calculate the execution ratio for RS232 serial
 tz = ((float)xtal / frame_rate) / (float)xtal;
 tb = ((float)1.0 / serial_rx_baud);

 execratio_serial = (int)(tz/tb);

// set the maximum allowed ratio for serial
 if (execratio_serial > 40)
    execratio_serial = 40;
}

//==============================================================================
// PIO read - Port function
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16		register data
//==============================================================================
UINT16 pio_r (UINT16 port, struct z80PortRead *port_s)
{
#if 0
 printf("pio_r port 0x%02x\n", port);
#endif

 if (debuglvl > 0)
    fprintf(debuglog, "pio_r port 0x%02x\n", port);

 switch (port & 0x03)
    {
     case 0x00 : // PIO port A data port
                 return pio_a.data;
     case 0x01 : // PIO port A control port
#ifdef use_pioinfo
                 printf("pio_r: port A control = 0x%02x\n", pio_a.cont);
#endif                 
                 return pio_a.cont;
     case 0x02 : // PIO port B data port
                 if ((key_tape) && (tape_i != NULL)) // if we have a file to use
                    pio_b.data = (pio_b.data & B8(11111110)) | tape_r();
                 if (coms1 != -1)   // if we have serial
                    pio_b.data = (pio_b.data & B8(11100111)) | serial_r();
                 else
                    pio_b.data = (pio_b.data & B8(11100111)) | B8(00001000); // stop bit, CTS
                 return pio_b.data;
     case 0x03 : // PIO port B control port
#ifdef use_pioinfo
                 printf("pio_r: port B control = 0x%02x\n", pio_b.cont);
#endif
                 return pio_b.cont;
    } 
 return 0xFFFF;
}

//==============================================================================
// PIO write - Port function
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void pio_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
 if (debuglvl > 0)
    fprintf(debuglog, "pio_w port 0x%02x = 0x%02x\n", port, data);

 switch (port & 0x03)
    {
     case 0x00 : // PIO port A data port
                 pio_a.data = data;
                 if ((print_a != NULL) || (print_b != NULL))  // if printer out
                    printer_w(data);
                 break;
     case 0x01 : // PIO port A control port
                 pio_control(&pio_a, data, 'A');
                 break;
     case 0x02 : // PIO port B data port
                 pio_b.data = data;
#ifdef use_audio
                 sound_w(data);
#endif
                 if (tape_o != NULL)	// if tape out
                    tape_w(data);
                 if (coms1 != -1)   	// if serial
                    serial_w(data);   
                 break;
     case 0x03 : // PIO port B control port
                 pio_control(&pio_b, data, 'B');
                 break;
    }             
}
