//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                              uBee512 module                                *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Provides init, application loop, and deinit.
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.3.0 - 1 September 2007, S.J.Kay
// - Added PIO polling and PIO configure (for interrupts) with execution
//   ratio change feature when interrupts are made active.
// - Added RS232 serial TX and RX options to set communication port,  baud
//   rates for each direction, data and stop bits: --coms, --baud, --baudtx,
//   --baudrx, --datab and --stopb.
// - Added printer output options --print and --printa.
// - Additional directory 'printer' created in the user's home account in the
//   .ubee512 directory.
// - Added --tapevol option for setting volume level of tape wave file.
// - Changed some error messages to put out function name first.
//
// v1.2.0 - 21 August 2007, S.J.Kay
// - Added --tapeo, --tapei, and --tapesamp options to support tape functions.
// - Additional directory 'tapes' created in the user's home account in the
//   .ubee512 directory.
// - Added function module calls for init, deinit and reset.
// - disk_and_roms() function name changed to set_account_paths()
//
// v1.1.0 - 18 August 2007, S.J.Kay
//   Code has been modified to allow the XTAL frequency to be specified.  In
//   theory if the host PC can emulate the Z80 code at that speed then that
//   is what you will get with a frame rate of (n) FPS.
//
//   See the README file for an explanation of how the speed emulation works.
//
// - stdout and stderr on windows now directed back to the console instead
//   of stdout.txt and stderr.txt files.
// - Modified the cursor blinking and Premium hardware inverse/flashing
//   video timer methods depending on if turbo mode is used.  This keeps the
//   flash rate correct.
// - Added frate command line option, mainly intended for hacking.
// - Added vblank command line option, mainly intended for use with the
//   turbo option to make the host timer determine the 50Hz VBLANK rate.
// - Added XTAL command line option (-x) where the xtal rate in MHz can be
//   specified in floating point format.
// - Added sound command line options: off, prop and normal.
// - Added volume setting command line option.
// - Changed sound hold-off to use a timer instead of Z80 cycle counter.
// - Changed from getopt to getopt_long command arguments. Uses same single
//   options as before.
// - Drive B not responding if a drive D was specified.  Problem caused by
//   floppyd[0] incorrectly labeled as floppyb[0] in the disk_and_roms()
//   function clearing any values already set for floppyb.
//
// v1.0.1 - 11 August 2007, S.J.Kay
// - Wrong default disk image type set for boot. Was boot.dip,  changed to
//   boot.dsk
//
// v1.0.0 - 8 August 2007, S.J.Kay
// - A major recoding of most of the original source to improve speed
//   regulation and for the addition of new features.
// - Create directories in users home accounts for the placement of disk and
//   ROM images for first time use.
// - Z80_cycles and associated variables now use unsigned 64 bit variables.
// - Added signal call back handler to allow control from other applications.
// - Added video timers and function calls in application loop for blinking
//   cursor and alpha+ hardware video flashing.
// - Added some code in the application loop to support the new sound emulation
//   feature.
// - Recoded the Z80 CPU speed regulation for the clock target.  The code now
//   uses the gettimeofday() function for unices and clock() function in mingw
//   builds to determine delay periods between mz80exec(n) function calls.
// - Added command line option for standard and standard+colour emulation.
// - Added command line monitor type options.
// - Added command line parameters for disk images for drives A-D.
// - Added turbo command line option.
// - Added fullscreen command line option.
// - Report init failures, not successes.
//
// v0.0.0 - 16 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//============================================================================== 
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
//==============================================================================   

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <getopt.h>
#include <string.h>
#include <time.h>
#include <SDL.h>

#ifdef MINGW
#include <conio.h>
#else
#include <sys/types.h> 		// various type definitions, like pid_t
#include <sys/time.h>
#include <sys/stat.h>
#include <signal.h>		// signal name macros, and the signal() prototype
#endif

#include "ubee512.h"
#include "z80.h"
#include "vdu.h"
#include "crtc.h"
#include "memmap.h"
#include "keyb.h"
#include "fdc.h"
#include "mz80.h"
#include "gui.h"
#include "paknet.h"
#include "pio.h"
#include "sound.h"
#include "tape.h"
#include "serial.h"
#include "function.h"

static int init(void);
static int deinit(void);
static int reset(void);

//==============================================================================
// structures and variables
//==============================================================================
FILE *debuglog;
int  debuglvl;

uint64_t z80_cycles;

int key_quit;
int key_reset;

int fullscreen;
int standard;

int flashvideo;

int done;
int loop;

#ifdef MINGW
clock_t ticks1, ticks2;
#else
static struct timeval tod;
uint64_t ticks_tod;
#endif

int delay;

char userhome[512];
char userhome_diskpath[512];
char userhome_romspath[512];
char userhome_tapepath[512];
char userhome_prnapath[512];
char userhome_prnbpath[512];

char romimage1[512];
char romimage2[512]; 
char romimage3[512];
char romimage4[512]; 

char floppya[512];
char floppyb[512];
char floppyc[512];
char floppyd[512];

char tapei[512];
char tapeo[512];

char printa[512];
char printb[512];

char coms1dev[512];

#ifdef use_audio
clock_t ticksound;
#endif

volatile char gui_signal;

int sound = 1;			// sound on by default
static int turbo;

char *floppy_a = NULL;
char *floppy_b = NULL;
char *floppy_c = NULL;
char *floppy_d = NULL;

char *tape_i = NULL;
char *tape_o = NULL;
static int tape_samples=TAPE_SAMPLE_FREQ;
static int tape_volume=TAPE_VOLUME;

char *print_a = NULL;
char *print_b = NULL;

char *coms1_dev = NULL;
int serial_tx_baud = SERIAL_TX_BAUD;
int serial_rx_baud = SERIAL_RX_BAUD;
int serial_databits = SERIAL_DATABITS;
int serial_stopbits = SERIAL_STOPBITS;

static float opt_xtal = XTALFREQ; // default XTAL frequency 
int frame_rate = FRAMERATE;	// default frame rate
extern int audio_volume;
int vblank_method = 0;
static char *opt_sound = NULL;
char *opt_monitor = NULL;
char *opt_colour = NULL;

extern int extendram;
extern int crtcresized;

extern int cur_blink;
extern int cur_mode;
extern int cur_pos;

extern int audio_silence;
extern int dobuffering;

int xtal;
int z80ms;

int z80_ratio;			// current z80 execution ratio (default = 1)

static int z80_block_cycles_def; // default number of Z80 cycles in 1 block
static int z80_blocks_def;	 // default number of Z80 blocks

static int z80_block_cycles_cur; // current number of Z80 cycles in 1 block
static int z80_blocks_cur;	// current number of Z80 blocks

int z80_block_cycles;		// working number of Z80 cycles in 1 block
int z80_blocks;			// working number of Z80 blocks

//==============================================================================
// Search an array of strings for the first occurence of the passed search
// string.  The string array must be be terminated by an empty string. 
//
//   pass: char *strg_array	pointer to a string array to be checked
//         char *strg_find	pointer to a string to search for
// return: int			1 if found, else 0
//==============================================================================
int string_search (char *strg_array[], char *strg_find)
{
 int i;
 
 for (i=0; strg_array[i][0] != 0; i++)
    {
     if (strcmp(strg_find, strg_array[i]) == 0)
        return 1;
    }
 return 0;
}        

#ifdef MINGW
#else
//==============================================================================
// External GUI signal handler.
//
// This allows an optional external GUI program to change values during the
// running of the emulator.  The code here is kept as simple as possible and
// simply just sets a flag to indicate that a signal was received.  The main
// loop checks the flag value to see if any action is required.
//
// Can be tested in unices by using the following to cause a reset():
// $kill -s 30 pid
//
// To find the pid use:
// $ps ax | grep ubee512
//
//   pass: int sig_num
// return: void
//==============================================================================
void signal_handler (int sig_num)
{
 gui_signal = 1;
}
#endif

//==============================================================================
// Create user home account directories and determine paths to be for disks,
// ROMs, tapes, and printers.
//
//   pass: void
// return: int				0 if no errors, else -1
//==============================================================================
int set_account_paths (void)
{
#if MINGW
#define SLASHCHAR '\\'
 strcat(userhome_diskpath, "c:\\ubee512\\disks\\");
 strcat(userhome_romspath, "c:\\ubee512\\roms\\");
 strcat(userhome_tapepath, "c:\\ubee512\\tapes\\");
 strcat(userhome_prnapath, "c:\\ubee512\\printer\\");
 strcat(userhome_prnbpath, "c:\\ubee512\\printer\\");
#else
#define SLASHCHAR '/'
 char *s;
 int dirres;
 mode_t mode;
 
 s = getenv("HOME");
 if (s == NULL)
    {
     printf("uBee512: Unable to get the user home path\n");
     return -1;
    }
 else
    strcat(userhome, s);

 strcat(userhome, "/.ubee512"); 

 strcat(userhome_diskpath, userhome); 
 strcat(userhome_diskpath, "/disks/");

 strcat(userhome_romspath, userhome);
 strcat(userhome_romspath, "/roms/");

 strcat(userhome_tapepath, userhome);
 strcat(userhome_tapepath, "/tapes/");

 strcat(userhome_prnapath, userhome);
 strcat(userhome_prnapath, "/printer/");

 strcat(userhome_prnbpath, userhome);
 strcat(userhome_prnbpath, "/printer/");

 mode = S_IRWXG | S_IRWXU | S_IROTH | S_IXOTH;
 
 dirres = mkdir(userhome, mode);
 if (dirres == 0)	// if there was no directory and it was created ok
    {
     dirres = mkdir(userhome_diskpath, mode);
     dirres = mkdir(userhome_romspath, mode);
     dirres = mkdir(userhome_tapepath, mode);
     dirres = mkdir(userhome_prnapath, mode);
     dirres = mkdir(userhome_prnbpath, mode);     
     
     printf("\n");
     printf("=========================================================================\n");
     printf("This is the first time you have run the ubee512 emulator in this account.\n");
     printf("Directories have been created in your home account for first time use.\n");
     printf("\n");
     printf("You now need to copy ROMs charrom.bin and rom1.bin into:\n");
     printf("%s\n", userhome_romspath);
     printf("and disk image(s) to:\n");
     printf("%s\n", userhome_diskpath);     
     printf("\n");     
     printf("When you have done this run the ubee512 emulator again.\n");          
     printf("=========================================================================\n");
     exit(0);		// exit so user can copy the required files
    } 

#endif

// Image A image
 if (floppy_a != NULL)
    {
     if (strchr(floppy_a, SLASHCHAR) == NULL)
        {
         floppya[0] = 0;
         strcat(floppya, userhome_diskpath);
         strcat(floppya, floppy_a); 
        }
     else    
        strcat(floppya, floppy_a);
     printf("Image A: %s\n", floppya);        
    }
 else
    {
     floppya[0] = 0; 
     strcat(floppya, userhome_diskpath);
     strcat(floppya, "boot.dsk");
     printf("Image A: %s\n", floppya);        
    } 

// Image B image
 if (floppy_b != NULL)
    {
     if (strchr(floppy_b, SLASHCHAR) == NULL)
        {
         floppyb[0] = 0;
         strcat(floppyb, userhome_diskpath);      
         strcat(floppyb, floppy_b); 
        }
     else    
        strcat(floppyb, floppy_b);
     printf("Image B: %s\n", floppyb);        
    }

// Image C image
 if (floppy_c != NULL)
    {
     if (strchr(floppy_c, SLASHCHAR) == NULL)
        {
         floppyc[0] = 0;
         strcat(floppyc, userhome_diskpath);      
         strcat(floppyc, floppy_c); 
        }
     else    
        strcat(floppyc, floppy_c);
     printf("Image C: %s\n", floppyc);        
    }

// Image D image
 if (floppy_d != NULL)
    {
     if (strchr(floppy_d, SLASHCHAR) == NULL)
        {
         floppyd[0] = 0;
         strcat(floppyd, userhome_diskpath);      
         strcat(floppyd, floppy_d); 
        }
     else    
        strcat(floppyd, floppy_d);
     printf("Image D: %s\n", floppyd);
    }

// Tape input file
 if (tape_i != NULL)
    {
     if (strchr(tape_i, SLASHCHAR) == NULL)
        {
         tapei[0] = 0;
         strcat(tapei, userhome_tapepath);      
         strcat(tapei, tape_i); 
        }
     else    
        strcat(tapei, tape_i);
     printf("Tape in file: %s\n", tapei);
    }

// Tape output file
 if (tape_o != NULL)
    {
     if (strchr(tape_o, SLASHCHAR) == NULL)
        {
         tapeo[0] = 0;
         strcat(tapeo, userhome_tapepath);      
         strcat(tapeo, tape_o); 
        }
     else    
        strcat(tapeo, tape_o);
     printf("Tape out file: %s\n", tapeo);
    }

 if ((tape_o != NULL) && (tape_i != NULL))
    if (strcmp(tapeo, tapei) == 0)
       {
        printf("uBee512: tape in and out can not be the same file\n");
        return -1;
       }

// Printer (ASCII) output file
 if (print_a != NULL)
    {
     if (strchr(print_a, SLASHCHAR) == NULL)
        {
         printa[0] = 0;
         strcat(printa, userhome_prnapath);      
         strcat(printa, print_a); 
        }
     else    
        strcat(printa, print_a);
     printf("Printer (ASCII) file: %s\n", printa);
    }

// Printer (BINIARY) output file
 if (print_b != NULL)
    {
     if (strchr(print_b, SLASHCHAR) == NULL)
        {
         printb[0] = 0;
         strcat(printb, userhome_prnapath);      
         strcat(printb, print_b); 
        }
     else    
        strcat(printb, print_b);
     printf("Printer (BINARY) file: %s\n", printb);
    }

// Serial communications device #1
 if (coms1_dev != NULL)
    {
     strcat(coms1dev, coms1_dev);
     printf("Serial communications dev: %s\n", coms1dev);
    }

// ROMs 
 strcat(romimage1, userhome_romspath); 
 strcat(romimage1, "rom1.bin"); 

 strcat(romimage2, userhome_romspath); 
 strcat(romimage2, "rom2.bin"); 

 strcat(romimage3, userhome_romspath); 
 strcat(romimage3, "rom3.bin"); 

 strcat(romimage4, userhome_romspath); 
 strcat(romimage4, "charrom.bin"); 

 return 0;
}

//==============================================================================
// Initialise.
//
//   pass: void
// return: int				0 if no errors, else -1
//==============================================================================
static int init (void)
{
 SDL_Surface *icon;

 printf(TITLESTRING"\n");

 if (set_account_paths() != 0)
    return -1;
 
#ifdef MINGW
#else
// set the GUI signal handler
 signal(30, signal_handler);
#endif

 if (debuglvl > 0)
    {
     debuglog = fopen(LOGFILE, "w"); 
     if (debuglog == NULL)
        {
         printf("init: Unable to write to log file\n");
         return -1;
        }
     fprintf(debuglog, "debug level = %i\n", debuglvl);
    }

 if (SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_TIMER) != 0)
    {
     printf("init: Failed SDL_Init - %s\n", SDL_GetError());
     return -1;
    }

 SDL_EventState(SDL_ACTIVEEVENT | SDL_KEYUP | SDL_KEYDOWN | SDL_SYSWMEVENT, SDL_IGNORE);

 SDL_WM_SetCaption(TITLESTRING, ICONSTRING);

 if ((icon = SDL_LoadBMP("uBee512.bmp")) != NULL)
    {
     SDL_WM_SetIcon(icon, NULL);
     SDL_FreeSurface(icon);
    }

 if (gui_init() != 0)
    {
     printf("init: Failed gui_init\n");    
     return -1;
    } 
 else
    {
    }   

 if (z80_init() != 0)
    {
     printf("init: Failed z80_init\n");    
     return -1;
    } 

 if (memmap_init() != 0)
    {
     printf("init: Failed memmap_init\n");    
     return -1;
    } 

 if (crtc_init() != 0)
    {
     printf("init: Failed crtc_init\n");
     return -1;
    }

 if (keyb_init() != 0)
    {
     printf("init: Failed keyb_init\n");
     return -1;
    } 

 if (vdu_init() != 0)
    {
     printf("init: Failed vdu_init\n");
     return -1;
    } 

 if (fdc_init() != 0)
    {
     printf("init: Failed fdc_init\n");
     return -1;
    } 

 if (pio_init() != 0)
    {
     printf("init: Failed pio_init\n");
     return -1;
    } 

 if (paknet_init() != 0)
    {
     printf("init: Failed paknet_init\n");
     return -1;
    } 

 if (function_init() != 0)
    {
     printf("init: Failed function_init\n");
     return -1;
    } 

 return 0;
}

//==============================================================================
// De-initialise.
//
//   pass: void
// return: int			result of de-init functions
//==============================================================================
static int deinit(void)
{
 int res;

 if (debuglvl > 0)
    fclose(debuglog);

 res = function_deinit() || pio_deinit() || paknet_deinit() || fdc_deinit() ||
       vdu_deinit() || keyb_deinit() || crtc_deinit() || memmap_deinit() ||
       z80_deinit() || gui_deinit();

 SDL_Quit();

 return res;
}

//==============================================================================
// Reset the virtual Microbee.
//
// Make sure all reset functions clear variables that are constantly active
// values.
//
//   pass: void
// return: int			result of reset functions
//==============================================================================
static int reset (void)
{
 z80_cycles = 0;
 done = 0;
 loop = 0;

#ifdef use_resetinfo
 printf("ubee512: emulation reset\n");
#endif

#ifdef MINGW
 ticks1 = clock();
#else
 gettimeofday(&tod, NULL);          
 ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif
 
 return z80_reset() || memmap_reset() || crtc_reset() || keyb_reset() ||
        vdu_reset() || fdc_reset() || pio_reset() || paknet_reset() ||
        function_reset();
}

//==============================================================================
// Z80 execution ratio set.
//
// The z80 execute method may be changed to execute fewer Z80 cycles in one Z80
// block but with more blocks by changing the ratio of these two values. 
//
// This is intended to allow PIO interrupts to be serviced faster as each
// interrupt is only serviced once on each call to mz80exec().  The ratio
// used can be set back to default after interrupts have completed.
//
//   pass: int divider		use 1 for default values
// return: void
//==============================================================================
void z80_exec_ratio (int divider)
{
 if (divider == 1)
    {
     z80_block_cycles_cur = z80_block_cycles_def;
     z80_blocks_cur = z80_blocks_def;
    }
 else
    {
     z80_block_cycles_cur = z80_block_cycles_def / divider;
     z80_blocks_cur = divider;
    } 

 z80_ratio = divider;
}

//==============================================================================
// Application setup.
//
//   pass: void
// return: void
//==============================================================================
void ApplicationSetup (void)
{
 float speedadj;
 float z80ms_r;

// convert XTAL to an integer in Hz
 xtal = (int)(opt_xtal * 1E6);

// calculate the number of Z80 cycles executed in 1 frame 
 z80_block_cycles_def = xtal / frame_rate;

// execution time in mS needed for one lot of z80_block_cycles_def
 z80ms_r = ((float)z80_block_cycles_def / xtal) * 1000.0;

// calculate a speed adjustment based on a constant percentage
 speedadj = (float)z80ms * (SPEED_PCT_ADJ / 100.0);

// apply the adjustment value
 z80ms = (int)(z80ms_r - speedadj);
 if (z80ms < 0)
    z80ms = 0;

 z80_blocks_def = 1;	// default blocks

 pio_configure(xtal, frame_rate, serial_rx_baud); // set exec ratios prota/serial
 
 z80_exec_ratio(1);	// use the values calculated here

 tape_config_out(xtal, tape_samples, tape_volume);
 tape_config_in(xtal);

 serial_config(xtal, serial_tx_baud, serial_rx_baud, serial_databits, serial_stopbits);

 sound_clock(xtal, SOUND_PROP); // sound proportional to XTAL freq

 if (opt_sound != NULL)
    {
     if (strcmp(opt_sound, "off") == 0)
        sound = 0;
     else
        if (strcmp(opt_sound, "prop") == 0)
           {
            sound = 1;
            sound_clock(xtal, SOUND_PROP); // sound proportional to XTAL freq
           } 
        else
           if (strcmp(opt_sound, "normal") == 0)
              {
               sound = 2;
               sound_clock(xtal, SOUND_3375); // sound set to normal rate
              }
    }             

 crtc_clock(xtal, vblank_method);  // set VBLANK method

 reset();			// resets the virtual Microbee

#ifdef MINGW
 ticks1 = clock();   
#else
 uint64_t ticks_tod;
 gettimeofday(&tod, NULL); 
 ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif 
}

//==============================================================================
// Application loop.
//
// Video timers
// ------------
// The CRTC blinking cursor and Alpha+ hardware inverse and flashing video
// timers use the accumulated Z80 cycle count,  this works well as the Z80 CPU
// execution is regulated with delays and so the instruction count is accurate
// over a period of time,  the rate will increase in turbo mode.
//
// Invalid instruction
// -------------------
// There appears to be a bug in v3.4 of makez80 or an uninitialised area of
// memory in another module.  It consistently occurs at 0xdec1 after
// 12,210,520 cycles are executed and a reset is then needed to fix the
// situation. The error occurs regardless of what CP/M OS is booted, the
// error still occurs when filling the boot track with 0s and a "jr $". Have
// found this to be a problem in win32 build but may be present on other
// builds too. There may be a variable not initilised correctly and only
// showing up in the win32 build. an old v2.7 of this code appears OK in
// this area but has other problems in itself.  The work around for now is
// to just reset and restart on each invalid instruction encountered.  This
// is NOT the end solution as some un-intened z80 instructions could get
// executed and may not even get an invalid instruction. It was found that
// the minimum work around required the following but decided to do a full
// reset instead. The minimum work around should give some clues.  A bank
// select issue could be causing rubbish to be exectued.
//
// z80_cycles = 0;   
// z80_reset();
// memmap_reset();
//
//   pass: void
// return: void
//==============================================================================
void ApplicationLoop (void)
{
 int dwResult;
 struct mz80context z80; 

 int cur_blink_rate;
 int cur_blink_last=0;
 int flashvideo_last=0;
 int invalid = 0;
 
 while (! done)
    {
#ifdef MINGW
     clock_t ticks_s; 		// for sound hold-off period
     ticks1 = clock();
#else
     uint64_t ticks_tod_snd;    // for sound hold-off period
     struct timeval tod_snd;
     struct timeval tod_x;     
     gettimeofday(&tod, NULL);          
     ticks_tod = (uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000));
#endif

// looping feature for speeding up PIO interrupt responses by having a
// smaller z80_block_cycles value.  Normally z80_blocks is set to 1 with
// z80_block_cycles equal to the default calculated rate.

     z80_blocks = z80_blocks_cur;
     z80_block_cycles = z80_block_cycles_cur;
   
     while (z80_blocks--)
        {
// Execute a block (Z80CYCLES) of Z80 instructions and return the result.
// The Z80CYCLES value used has been calculated for timing purposes.
         dwResult = mz80exec(z80_block_cycles);

// If the result is not 0x80000000, it's an address where an invalid
// instruction was hit.
         if (dwResult != 0x80000000)
            {
             mz80GetContext(&z80);
             if (++invalid > 1)
                {
                 printf("Invalid instruction at: 0x%4x\n", (unsigned int)z80.z80pc);
                 printf("           Clock ticks: %ld\n", z80.z80clockticks);
                } 
             z80_cycles = 0;   
             reset();
            }

         z80_cycles += mz80GetElapsedTicks(1);

// poll the PIO for interrupt events
         pio_polling();
        } 

// Resize the video when certain CRTC registers have been changed under Z80
// emulation. A wait period is introduced here to prevent the display being
// resized too often in a short time period.
     if (crtcresized)
        {
         if (crtcresized++ == 2)	// will be the 2nd time around loop
            crtc_videochange();		// resets crtcresized
        }

// Determine the current status for the CRTC blinking cursor and refresh it
// if this has changed.  The method used here depends on if turbo mode is
// used. if turbo mode is used then Z80 execution speed will not be known as
// no delays will be inserted, if not turbo then the rate must be determined
// by the Z80 cycle count to achieve smooth results.
     if (cur_mode > 1)			// only if blinking cursor
        {
         if (turbo)
            {
             if (cur_mode == 2)
                cur_blink_rate = 500;	// blinking at 1/32 field rate
             else   
                cur_blink_rate = 250;	// blinking at 1/16 field rate
#ifdef MINGW
             if ((clock() / cur_blink_rate) & 0x01)
#else
             gettimeofday(&tod_x, NULL);          
             if (((tod_x.tv_usec / 1000) / cur_blink_rate) & 0x01) // rolls over each sec
#endif
                cur_blink = 0xFF;
             else
                cur_blink = 0x00;
            }
         else
            {
             if (cur_mode == 2)
                cur_blink_rate = xtal / 2;	// blinking at 1/32 field rate
             else   
                cur_blink_rate = xtal / 4;	// blinking at 1/16 field rate        
             if ((z80_cycles / cur_blink_rate) & 0x01)
                cur_blink = 0xFF;
             else
                cur_blink = 0x00;
            }            
         if (cur_blink != cur_blink_last)
            {
             cur_blink_last = cur_blink;            
             crtc_redraw_char(cur_pos);
            }
        }

// Determine the current state of the alpha+ flashing video and refresh it
// if this has changed.  the method used here is the same as that used by
// the blinking cursor timer.
     if (extendram)			// only if extended RAM selected
        {
         if (turbo)
            {
#ifdef MINGW
             if ((clock() / 250) & 0x01)
#else
             gettimeofday(&tod_x, NULL);          
             if (((tod_x.tv_usec / 1000) / 250) & 0x01) // rolls over each sec
#endif
               flashvideo = 1;
            else
               flashvideo = 0;
            }
         else
            {
             if ((z80_cycles / (xtal / 4)) & 0x01)
                flashvideo = 1;
             else
                flashvideo = 0;
            }
         if (flashvideo != flashvideo_last)
            {
             flashvideo_last = flashvideo;            
             crtc_set_redraw();	// force crtc to redraw screen
            }
        }    

#if use_audio
// Sound buffering.  This is required as CPU emulation speed is affected by
// other processes occuring at the same time, we need to try to ensure the
// sound is slightly in front so that it does not matter if the Z80 lags
// behind slightly in creating the sound.  A 100mS delay has been found
// do be about right.  If the buffer delay is too long the lag will be
// noticable by the user.  If too short then break up, and crackling will be
// evident.
     if (dobuffering)
        {
         if (ticksound == -1)		// if just starting the buffering
#ifdef MINGW
            {
             ticksound = 0;
             ticks_s = clock();
            } 
#else
            { 
             ticksound = 0;            
             gettimeofday(&tod_snd, NULL);          
             ticks_tod_snd = (uint64_t)((tod_snd.tv_sec * 1000) + (tod_snd.tv_usec / 1000));
            }
#endif
         else
            {
#ifdef MINGW
             if ((clock() - ticks_s) >= SOUND_HOLD_OFF)
#else
             gettimeofday(&tod_snd, NULL);          
             if ((((uint64_t)((tod_snd.tv_sec * 1000) + (tod_snd.tv_usec / 1000))) - ticks_tod_snd) >= SOUND_HOLD_OFF)
#endif
                sound_unsilence();	// allow sound to play
            }
        }    
#endif

// Update the display (only if it's update flag is set)           
     gui_redraw();

// Check keys,  if the keys do not get checked in the Z80 code then we must make
// sure it gets checked periodically for the emulator functionality keys.
     keyb_keysupdate();

// Handle external GUI signals
     if (gui_signal)
        {
         reset();
         gui_signal = 0;
        }

// Check for exit and reset key flags
     if (key_quit)
        {
         key_quit = 0;
         done = 1;
        } 
     if (key_reset)
        {
         key_reset = 0;
         reset();
        }

// Insert delays to achieve correct emulated Z80 CPU speed.  The Z80 speed
// may be made a little faster to help prevent audio break up.
//
// The MINGW build relies on a 1mS timer (1ms granularity) and the delay
// required is calculated from the last block (Z80CYCLES) of Z80
// instructions.
//
// The non MINGW builds (unices) uses the gettimeofday function and this is
// converted to a 64 bit variable with a 1mS resolution.
//
     if (turbo)
        SDL_Delay(0);
     else
        {
#ifdef MINGW
// Find the execution delay required for the last (1/frame_rate)
         ticks2 = clock() - ticks1;
         if (ticks2 >= 0)		// if clock() has not wrapped around
            {                   	// else use the last delay value
             delay = z80ms - ticks2;
            }
#else
         gettimeofday(&tod, NULL);
         delay = z80ms - ((uint64_t)((tod.tv_sec * 1000) + (tod.tv_usec / 1000)) - ticks_tod);
#endif
         
// Execute the delay period.  If a delay is not required then a delay of 0
// is used to give up the CPU for other processes.
         if (delay < 0)
            SDL_Delay(0);
         else
            SDL_Delay(delay);
        }
    }    
}

//==============================================================================
// Application usage
//
//   pass: void
// return: void
//==============================================================================
void ApplicationUsage (void)
{
 char const usage[] =
TITLESTRING"\n"
"\n"
"Usage: ubee512 [options]\n"
"\n"
"-a, --image_a=file      Image file path & filename for floppy drive A\n"
"-b, --image_b=file      Image file path & filename for floppy drive B\n"
"-c, --image_c=file      Image file path & filename for floppy drive C\n"
"-d, --image_d=file      Image file path & filename for floppy drive D\n"
"\n"
"                        If you specify a path as part of the image name then\n"
"                        it will be accessed from that area.  If only an\n"
"                        image name is specified the emulator will look for\n"
"                        it here:\n"
"\n"
"                        Windows:\n"
"                        c:\\ubee512\\disks\\\n"
"\n"
"                        Unices:\n"
"                        /home/username/.ubee512/disks/\n"
"\n"
"                        Hint: if you want to access an image in the current\n"
"                        directory you need to force it using the following\n"
"                        method:\n"
"\n"
"                        Windows:\n"
"                        ubee512 -a .\\myimage.dsk\n"
"\n"
"                        Unices:\n"
"                        ubee512 -a ./myimage.dsk\n"
"\n"
"--baud=rate             Set serial communications baud rate for both TX and\n"
"                        RX.  A value from 1 to 19200 is allowed.  Default\n"
"                        rate is 300 baud.  If Individual baud rates are\n"
"                        required for TX and RX then use the --baudtx and\n"
"                        --baudrx options instead.  This value must match the\n"
"                        Microbee serial application's value.\n"
"\n"
"--baudtx=rate           Set serial communications baud rate for TX only.  A\n"
"                        value from 1 to 19200 is allowed.  Default rate is\n"
"                        300 baud.  This value must match the Microbee serial\n"
"                        application's value.\n"
"\n"
"--baudrx=rate           Set serial communications baud rate for RX only.  A\n"
"                        value from 1 to 19200 is allowed.  Default rate is\n"
"                        300 baud. This value must match the Microbee serial\n"
"                        application's value.\n"
"\n"
"--coms=port             Serial communications port for emulation of RS232.\n"
"                        On Unices specify a device, on Windows specify the\n"
"                        com port number.  No serial communications will be\n"
"                        emulated if this option is not specified.\n"
"\n"
"--datab=bits            Set serial communications number of data bits.  A\n"
"                        value from 5 to 8 is allowed.  Default value is 8\n"
"                        data bits.  This value must match the Microbee\n"
"                        serial application's value.\n"
"\n"
"--frate=fps             Frame rate,  the default is 30 FPS,  an integer value\n"
"                        between 1 and 1,000,000 is allowed.  This option is\n"
"                        intended for 'hacking' and code development use.\n"
"\n"
"-f, --fullscreen        Start in full screen mode (use ALT+ENTER to toggle).\n"
"\n"
"-h, --help, --usage     Display help information on command line usage\n"
"\n"
"-m, --monitor=type      Monitor type, if this option is not specified a\n"
"                        colour monitor is the default. <type> may be one\n"
"                        of the following:\n"
"\n"
"                        a : amber monitor\n"
"                        g : green monitor\n"
"                        w : white monitor, white foreground on black\n"
"                            background\n"
"                        b : black monitor, black foreground on white\n"
"                            background\n"
"                        c : colour monitor\n"
"\n"
"                        Note: This option by itself does not force the\n"
"                        emulation into a standard model Microbee, it's use\n"
"                        simply determines what sort of monitor is connected\n"
"                        to the emulated Microbee, be it a standard or\n"
"                        Premium model.  If you want to force the Premium\n"
"                        graphics 'off' you must uses the -s<col> option.  If\n"
"                        <col> = m then this option can then further set\n"
"                        the desired monochrome colour.\n"
"\n"
"--print=file            Printer output to file,  the output is not modified.\n"
"                        The location of the file follows the same rule as\n"
"                        used for the drive image options except default\n"
"                        directory is 'printer'.\n"
"\n"
"--printa=file           Printer output to file,  the output is converted to\n"
"                        ASCII decimal. The location of the file follows the\n"
"                        same rule as used for the drive image options except\n"
"                        default directory is 'printer'.\n"
"\n"
"--sound=method          Determine the method used for sound,  the default\n"
"                        method is 'prop':\n"
"\n"
"                           off : sound is turned off\n"
"                          prop : sound is proportional to XTAL frequency\n"
"                        normal : sound rate forced as if 3.375 MHz XTAL\n"
"\n"
"-s, --standard=col      Standard graphics Microbee emulation,  turn off\n"
"                        the Premium (alpha+) graphics and use a standard\n"
"                        green monitor. The <col> specifier allows for\n"
"                        standard monochrome or colour support:\n"
"\n"
"                        c : standard Microbee with colour\n"
"                        m : standard Microbee with Green monochrome\n"
"\n"
"                        If using this option with the 'm' parameter then the\n"
"                        -m<mode> option can also be specified to select the\n"
"                        monochrome colour that is preferred.\n"
"\n"
"--stopb=bits            Set serial communications number of stop bits.  A\n"
"                        value from 1 to 2 is allowed.  Default value is 1\n"
"                        data bits.  This value must match the Microbee\n"
"                        serial application's value for TX.\n"
"\n"
"--tapei=file            Tape input from a WAV file.  The location of the file\n"
"                        follows the same rule as used for the drive image\n"
"                        options except default directory is 'tapes'.\n"
"\n"
"--tapeo=file            Tape output to a WAV file.  The location of the file\n"
"                        follows the same rule as used for the drive image\n"
"                        options except default directory is 'tapes'.\n"
"\n"
"--tapesamp=frequency    Tape output sample frequency in Hz.  Default is\n"
"                        22050 Hz.\n"
"\n"
"--tapevol=level         Tape output wave file volume level.  A level of 0 to\n"
"                        127 is allowed. (default level is 16)\n"
"\n"
"-t, --turbo             Turbo mode, executes Z80 code as fast as possible.\n"
"                        Without this option the emulation attempts to keep\n"
"                        Z80 CPU execution to match the XTAL value.  This\n"
"                        option is intended for 'hacking' and code development\n"
"                        use.  There are much faster methods if more speed is\n"
"                        required. (see the README file)\n"
"\n"
"--vblank=method         Vertical blanking method to be employed. This is\n"
"                        only intended for 'hacking' when experimenting with\n"
"                        turbo mode and/or high XTAL speeds.  It is not\n"
"                        required or even recommended to be used if 3.375 MHz\n"
"                        or 2 MHz emulation is desired.\n"
"\n"
"                        0 : 50 Hz VBLANK rate derived from Z80 cycles and is\n"
"                            proportional to the XTAL frequency.\n"
"                        1 : 50 Hz VBLANK rate derived from the host timer.\n"
"\n"
"                        When running in turbo mode then setting the <method>\n"
"                        equal to 1 will ensure that a VBLANK rate of 50Hz\n"
"                        will be used.  Without this, key repeating may be\n"
"                        too fast.\n"
"\n"
"--version               Obtain the version number.\n"
"\n"
"--vol=level             Set the sound volume level.  A level of 0 to 127 is\n"
"                        allowed. (default level is 16)\n"
"\n"
"-x, --xtal=frequency    Set the XTAL frequency for emulation in MHz.\n"
"                        Standard emulation frequencies are 3.375 and 2.0\n"
"                        MHz.  All other frequencies are classed as 'hacking'.\n" 
"                        3.375 MHz is the default XTAL frequency.  Frequency\n"
"                        is entered as a floating point number.\n"
"\n"
"                        The highest frequency possible is determined by the\n"
"                        host platform.  As the value increases to the point\n"
"                        where uBee512 can no longer regulate the Z80\n"
"                        execution rate the frame rate will decrease (slower\n"
"                        screen update periods).\n"
"\n"
"-z, --debug             Debugging output (for development purposes only)\n"
"                        writes to a log file.\n"
"\n"
"If you have any new feature suggestions, bug reports, etc. then please send\n"
"an email to: ubee512@gmail.com\n";

#ifdef MINGW
// make this process easier for Windows users.
 char ch;
 int i = 0;
 int lineswanted = 23;
 while (lineswanted)
    {
     ch = (usage[i]);
     
     if (ch != '\n')
        {
         printf("%c", ch);
         i++;
        }
     else
        {    
         i++;	// get past the \n character
         printf("\n");

         if (! usage[i])
            return;

         if (--lineswanted == 0)
            {                 
             printf("\r======= ENTER=next line ======= SPACE=next screen ======= ESC=done =======" );
             ch = getch();
             printf("\r                                                                          \r");             
             switch (ch)
                {
                 case 13 : lineswanted = 1;
                           break;
                 case 32 : lineswanted = 22;
                           break;
                 case 27 : lineswanted = 0;
                           break;                           
                }           
            }
        }
    }
#else
 printf("%s", usage);
#endif 
}

//==============================================================================
// Main program start point.
//
//   pass: int argc, char *argv[]	Command line arguments
// return: int				0 if no error, 1 if error
//==============================================================================
int main (int argc, char *argv[])
{
 static struct option long_options[] =
    {
     {"fullscreen", no_argument,       0, 'f'},
     {"help",       no_argument,       0, 'h'},
     {"turbo",      no_argument,       0, 't'},
     {"debug",      no_argument,       0, 'z'},
     {"version",    no_argument,       0,  3 },
     {"usage",      no_argument,       0, 'h'},

     {"coms",       required_argument, 0,  9 },
     {"baud",       required_argument, 0, 10 },
     {"baudtx",     required_argument, 0, 11 },
     {"baudrx",     required_argument, 0, 12 },
     {"datab",      required_argument, 0, 13 },
     {"stopb",      required_argument, 0, 14 },

     {"print",      required_argument, 0, 15 },
     {"printa",     required_argument, 0, 16 },          
     
     {"frate",      required_argument, 0,  4 },
     {"image_a",    required_argument, 0, 'a'},
     {"image_b",    required_argument, 0, 'b'},               
     {"image_c",    required_argument, 0, 'c'},
     {"image_d",    required_argument, 0, 'd'},

     {"monitor",    required_argument, 0, 'm'},
     {"sound",      required_argument, 0,  1 },
     {"standard",   required_argument, 0, 's'},

     {"tapei",      required_argument, 0,  5 },
     {"tapeo",      required_argument, 0,  6 },
     {"tapesamp",   required_argument, 0,  7 },
     {"tapevol",    required_argument, 0,  8 },     

     {"vblank",     required_argument, 0,  2 },
     {"vol",        required_argument, 0, 'v'},
     {"xtal",       required_argument, 0, 'x'},
     {0, 	    0,                 0,  0 }
    };

 char *monitor_args[] =
 {
  "a",
  "g",
  "w",
  "b",
  "c",
  ""
 };

 char *standard_args[] =
 {
  "c",
  "m",
  ""
 };

 char *sound_args[] =
 {
  "off",
  "prop",
  "normal",
  ""
 };

 int int_arg;
 float float_arg;
 
#define PARMERR_MESG "ubee512: option %s argument of '%s' is not permitted\n" 
#define TRY_MESG "ubee512: Try `ubee512 --help' or `ubee512 --usage' for more information.\n"
 
// don't want stdout.txt or stderr.txt files so put output back to console.
#ifdef MINGW
 freopen("CON", "w", stdout);
 freopen("CON", "w", stderr);
#endif

 int c;
     
 while (1)
    {
// getopt_long stores the option index here.
     int option_index = 0;
     
     c = getopt_long (argc, argv, "fhtza:b:c:d:m:s:v:x:", long_options, &option_index);
     
// detect the end of the options.
     if (c == -1)
        break;
     
     switch (c)
        {
         case  1 : if (string_search(sound_args, optarg))
                      opt_sound = optarg;
                   else
                      {
                       printf(PARMERR_MESG, "'--sound'", optarg);
                       exit(1);   
                      } 
                   break;                   
                   break;
         case  2 : int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 0) || (int_arg > 1))
                      {
                       printf(PARMERR_MESG, "'--vblank'", optarg);
                       exit(1);
                      }
                   else
                      vblank_method = int_arg;    
                   break;
         case  3 : printf("%s\n", APPVER);
                   exit(0);
                   break;                    
         case  4 : int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 1) || (int_arg > 1000000L))
                      {
                       printf(PARMERR_MESG, "'--frate'", optarg);
                       exit(1);
                      }
                   else
                      frame_rate = int_arg;    
                   break;
         case  5 : tape_i = optarg;
                   break;
         case  6 : tape_o = optarg;
                   break;                   
         case  7 : int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 1) || (int_arg > 1000000L))
                      {
                       printf(PARMERR_MESG, "'--tapesamp'", optarg);
                       exit(1);
                      }
                   else
                      tape_samples = int_arg;    
                   break;
         case  8 : int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 0) || (int_arg > 127))
                      {
                       printf(PARMERR_MESG, "'--tapevol'", optarg);
                       exit(1);
                      }
                   else
                      tape_volume = int_arg;
                   break;   
         case  9 : coms1_dev = optarg;
                   break;
         case 10 : int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 1) || (int_arg > 19200))
                      {
                       printf(PARMERR_MESG, "'--baud'", optarg);
                       exit(1);
                      }
                   else
                      {
                       serial_tx_baud = int_arg;
                       serial_rx_baud = int_arg;
                      } 
                   break;
         case 11 : int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 1) || (int_arg > 19200))
                      {
                       printf(PARMERR_MESG, "'--baudtx'", optarg);
                       exit(1);
                      }
                   else
                      serial_tx_baud = int_arg;
                   break;
         case 12 : int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 1) || (int_arg > 19200))
                      {
                       printf(PARMERR_MESG, "'--baudrx'", optarg);
                       exit(1);
                      }
                   else
                      serial_rx_baud = int_arg;
                   break;
         case 13 : int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 5) || (int_arg > 8))
                      {
                       printf(PARMERR_MESG, "'--datab'", optarg);
                       exit(1);
                      }
                   else
                      serial_databits = int_arg;
                   break;
         case 14 : int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 1) || (int_arg > 2))
                      {
                       printf(PARMERR_MESG, "'--stopb'", optarg);
                       exit(1);
                      }
                   else
                      serial_stopbits = int_arg;
                   break;
         case 15 : print_b = optarg;
                   break;
         case 16 : print_a = optarg;
                   break;                   
         case 'f': fullscreen++;
                   break;
         case 'h': ApplicationUsage();
                   exit(0);
                   break;
         case 't': turbo++;
                   break;
         case 'z': debuglvl++;
                   break;
         case 'a': floppy_a = optarg;
                   break;
         case 'b': floppy_b = optarg;
                   break;
         case 'c': floppy_c = optarg;
                   break;
         case 'd': floppy_d = optarg;
                   break;
         case 'm': if (string_search(monitor_args, optarg))
                      opt_monitor = optarg;
                   else
                      {
                       printf(PARMERR_MESG, "'-m, --monitor'", optarg);
                       exit(1);   
                      } 
                   break;
         case 's': if (string_search(standard_args, optarg))
                      {
                       standard = 1;
                       opt_colour = optarg;
                      } 
                   else
                      {
                       printf(PARMERR_MESG, "'-s, --standard'", optarg);
                       exit(1);   
                      } 
                   break;
         case 'v': int_arg = -1;
                   sscanf(optarg, "%d", &int_arg);
                   if ((int_arg < 0) || (int_arg > 127))
                      {
                       printf(PARMERR_MESG, "'--vol'", optarg);
                       exit(1);
                      }
                   else
                      audio_volume = int_arg;
                   break;
         case 'x': float_arg = -1.0;
                   sscanf(optarg, "%f", &float_arg);
                   if (float_arg <= 0.0)
                      {
                       printf(PARMERR_MESG, "'-x, --xtal'", optarg);
                       exit(1);
                      }
                   else
                      opt_xtal = float_arg;    
                   break;
         case '?': // getopt_long already printed an error message.
                   printf(TRY_MESG);
                   exit(1);
         default : exit(1);
        }
    }

// Check if any other arguments are incorrectly specified
 if (optind < argc)
    {
     printf("ubee512: %d additional arguments were specified that are not recognised:\n", argc - optind);         
     printf("ubee512: ");
     while (optind < argc)
        printf ("%s ", argv[optind++]);
     putchar ('\n');
     printf(TRY_MESG);
     exit(1);
    }

 if (init() != 0)
    {
     printf("main: Fatal error during initialisation.\n");
     SDL_Quit();
     exit(1);
    }

 ApplicationSetup();
 ApplicationLoop();
 
 if (deinit() != 0)
    {
     printf("main: Error while de-initialising.\n");
     exit(1);
    }

 return 0;
}
