//******************************************************************************
//*                                  uBee512                                   *
//*             An emulator for the Premium Microbee 512K FDD model            *
//*                                                                            *
//*                                 VDU module                                 *
//*                                                                            *
//*                       Copyright (C) 2007 Stewart Kay                       *
//******************************************************************************
//
// Emulates the graphics hardware for Standard and Premium models (Alpha+)
//
// - Character ROM
// - Screen RAM
// - PCG RAM
// - Colour RAM
// - Attribute RAM
//
//==============================================================================
// ChangeLog (most recent entries are at top)
//==============================================================================
// v1.0.0 - 21 June 2007, S.J.Kay
// - Added standard and standard+colour machine options to vdu_colcont_w and
//   vdu_lvdat_w functions.
// - Added vdu_lvdat_r and vdu_lvdat_w port handlers.
// - Implemented Alpha+ (Premium) graphics.
//
// v0.0.0 - 8 June 2007, S.J.Kay
// Start with "nanowasp" source distribution version 0.22. An emulator for the
// microbee 128k. Copyright (C) 2000-2003  David G. Churchill.

//==============================================================================
/*
 *  uBee512 - An emulator for the Premium Microbee 512K FDD model
 *  Copyright (C) 2007  Stewart Kay
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/ 
//==============================================================================

#include <stdio.h>
#include <string.h>
#include <SDL.h>

#include "vdu.h"
#include "crtc.h"
#include "z80.h"
#include "ubee512.h"
#include "memmap.h"
#include "mz80.h"

#include "macros.h"

//==============================================================================
// structures and variables
//==============================================================================
int latchrom;
int colour_cont;		// colour control port (0x08) value
int x_colour_cont = -1;
int colourram;			// colour RAM selected if bit 4=1 else PCG RAM
int x_colourram;

char charrom[CHARROMSIZE];

extern char romimage4[];

extern int mem_mode;
extern int standard;
extern int usecolour;

#ifdef use_alpha_p
// Alpha+ (Premium) variables, 8K for Screen, Colour and attribute, 32K for PCG
char screen_ram_ap[4 * 0x0800],
     colour_ram_ap[4 * 0x0800],
     attrib_ram_ap[4 * 0x0800],
     pcg_ram_ap[16 * 0x0800];

int videobank;			// LV DAT bank bits 0-3
int x_videobank;
int attribram;			// LV DAT attribute RAM selected if bit 4=1
int x_attribram;
int extendram;	        	// LV DAT extended graphics selected if bit 7=1
int x_extendram;
int lv_dat;			// LV DAT port (0x1c) value
int x_lv_dat = -1;

#else
char videoram[VIDEORAMSIZE],
     pcgram[PCGRAMSIZE];
#endif

//==============================================================================
// VDU initialise
//
//   pass: void
// return: int				0 if success else -1
//==============================================================================
int vdu_init (void)
{
 FILE *fp;

#ifdef use_alpha_p
// Alpha+ (Premium) variables, 8K for Screen, Colour and attribute, 32K for PCG
 memset(screen_ram_ap, 0, sizeof(screen_ram_ap));
 memset(colour_ram_ap, 0, sizeof(colour_ram_ap));
 memset(attrib_ram_ap, 0, sizeof(attrib_ram_ap));
 memset(pcg_ram_ap,    0, sizeof(pcg_ram_ap));

 lv_dat=0;		// LV DAT port (0x1c) value
 x_lv_dat=-1;
 videobank = 0;		// LV DATA bank bits 0-3
 attribram = 0;		// LV DATA attribute RAM selected if bit 4=1
 extendram = 0;	        // LV DATA extended graphics selected if bit 7=1

 colour_cont = 0;	// colour control port (0x08) value
 x_colour_cont = -1;
 colourram = 0;		// colour RAM selected if bit 6=1 else PCG RAM
#else
 memset(videoram, 0, VIDEORAMSIZE);
 memset(pcgram, 0, PCGRAMSIZE);
#endif

// load character ROM
 fp = fopen(romimage4, "rb");
 if (fp == NULL)
    {
     fprintf(stderr, "vdu_init: unable to read %s\n", romimage4);
     return -1;
    }

 if (fread(charrom, 1, CHARROMSIZE, fp) != CHARROMSIZE)
    {
     fprintf(stderr, "vdu_init: charrom.bin must be %i bytes\n", CHARROMSIZE);
     return -1;
    }

 fclose(fp);
 
 return 0;
}

//==============================================================================
// VDU de-initialise
//
//   pass: void
// return: int				0
//==============================================================================
int vdu_deinit (void)
{
 return 0;
}

//==============================================================================
// VDU reset
//
//   pass: void
// return: int				0
//==============================================================================
int vdu_reset (void)
{
 latchrom = 0;

#ifdef use_alpha_p
// Alpha+ (Premium) variables, 8K for Screen, Colour and attribute, 32K for PCG
 memset(screen_ram_ap, 0, sizeof(screen_ram_ap));
 memset(colour_ram_ap, 0, sizeof(colour_ram_ap));
 memset(attrib_ram_ap, 0, sizeof(attrib_ram_ap));
 memset(pcg_ram_ap,    0, sizeof(pcg_ram_ap));

// memcpy(z80mem+0xF000, &screen_ram_ap[0], 0x800);
// memcpy(z80mem+0xF800, &pcg_ram_ap[0], 0x800); 

 lv_dat=0;		// LV DAT port (0x1c) value
 x_lv_dat=-1;
 videobank = 0;		// LV DATA bank bits 0-3
 attribram = 0;		// LV DATA attribute RAM selected if bit 4=1
 extendram = 0;	        // LV DATA extended graphics selected if bit 7=1

 colour_cont = 0;	// colour control port (0x08) value
 x_colour_cont = -1;
 colourram = 0;		// colour RAM selected if bit 6=1 else PCG RAM
#endif

 return 0;
}

#ifdef use_alpha_p
//==============================================================================
// Switch video out.
//
// Must switch out the following video banks from 8000/8800 or F000/F800:
// 
// 1. Copy main memory to attribute/screen RAM (bit 4=1 if attribute) bank
//    (bits 0-3).
// 2. Copy main memory to PCG/Colour RAM bank (bits 0-3) to main memory. The
//    state of port 0x08 will determine if PCG or Colour RAM.
//
// NOTES:
// - if bit 7=0 then banks 0 still must be put back in to place for the above.
// - character ROM is never in main memory so no check is needed for it.
//
//   pass: void
// return: void
//==============================================================================
void vdu_switchvideo_out (void)
{
 int vb;
 int video;

 if (! (mem_mode & BANK_VRAM))		// if video RAM enabled
    {
     if (mem_mode & BANK_VADD)		// if video at 8000-8FFF
        video = 0x8000;
     else
        video = 0xF000;   

     if (! x_extendram)	        	// if not extended graphics selected
        vb = 0;
     else    
        vb = x_videobank;
       
     if (x_attribram)             
        memcpy(&attrib_ram_ap[(vb & B8(00000011)) * 0x0800], z80mem+video, 0x800);
     else
        memcpy(&screen_ram_ap[(vb & B8(00000011)) * 0x0800], z80mem+video, 0x800);
     if (x_colourram)
        memcpy(&colour_ram_ap[(vb & B8(00000011)) * 0x0800], z80mem+video+0x0800, 0x800);
     else
        memcpy(&pcg_ram_ap[vb * 0x0800], z80mem+video+0x0800, 0x800);
    } 
}

//==============================================================================
// Switch video in.
//
// Must switch in the following video banks to 8000/8800 or F000/F800:
// 
// 1. Copy attribute/screen RAM (bit 4=1 if attribute) bank (bits 0-3) to
//    main memory.
// 2. Copy PCG/Colour RAM bank (bits 0-3) to main memory. The state of port
//    0x08 will determine if PCG or Colour RAM 
//
// NOTES:
// - if bit 7=0 then banks 0 still must be put back in to place for the above.
// - character ROM is never in main memory so no check is needed for it.
//
//   pass: void
// return: void
//==============================================================================
void vdu_switchvideo_in (void)
{
 int vb;
 int video;

 if (! (mem_mode & BANK_VRAM))		// if video RAM enabled
    {
     if (mem_mode & BANK_VADD)		// if video at 8000-8FFF
        video = 0x8000;
     else
        video = 0xF000;   

     if (! extendram)	                // if not extended graphics selected
        vb = 0;
     else    
        vb = videobank;
   
     if (attribram)             
        memcpy(z80mem+video, &attrib_ram_ap[(vb & B8(00000011)) * 0x0800], 0x800);
     else
        memcpy(z80mem+video, &screen_ram_ap[(vb & B8(00000011)) * 0x0800], 0x800);
     if (colourram)             
        memcpy(z80mem+video+0x0800, &colour_ram_ap[(vb & B8(00000011)) * 0x0800], 0x800);
     else
        memcpy(z80mem+video+0x0800, &pcg_ram_ap[vb * 0x0800], 0x800);

// save the current video state
     x_lv_dat = lv_dat;			// port (0x1c) value
     x_videobank = videobank;		// bank bits 0-3
     x_attribram = attribram;		// attribute RAM select
     x_extendram = extendram; 		// extended graphics select
     x_colour_cont = colour_cont;
     x_colourram = colourram;
    } 
}
#endif

//==============================================================================
// Video memory read.
//
// Note: If the 'standard' option is in affect then this code should behave like
// a standard machine. (extenram=0, attribram=0)
//
//   pass: UINT32 addr
//         struct MemoryReadByte *mem_s
// return: UINT8
//==============================================================================
UINT8 vdu_vidmem_r (UINT32 addr, struct MemoryReadByte *mem_s)
{
// if (debuglvl > 0)
//    fprintf(debuglog, "vdu_vidmem_r at 0x%04x\n", (unsigned int)addr);

#ifdef use_alpha_p
 int vb;

 if (! extendram)		// if the alpha+ graphics is not enabled
    vb = 0;
 else    
    vb = videobank;
    
 if (addr & 0x0800)
    {
     if (! colourram)
        return pcg_ram_ap[vb * 0x0800 + (addr & 0x07FF)];
     else   
        return colour_ram_ap[(vb & B8(00000011)) * 0x0800 + (addr & 0x07FF)];
    }    
 else
    {
     if (! latchrom)
        {
         if (! attribram)             
            return screen_ram_ap[(vb & B8(00000011)) * 0x0800 + (addr & 0x07FF)];
         else   
            return attrib_ram_ap[(vb & B8(00000011)) * 0x0800 + (addr & 0x07FF)];
        }
     else       
/* FIXME this will not work if disp_start is set to just before 0x2000, and the
   memory addresses would overlap 0x2000 */
        return charrom[((disp_start & 0x2000) >> 2) + (addr & 0x07FF)];
    }
#else
 if (addr & 0x0800)
    return pcgram[addr & 0x07FF];
 else
    {
     if (! latchrom)
        return videoram[addr & 0x07FF];
     else
/* FIXME this will not work if disp_start is set to just before 0x2000, and the
   memory addresses would overlap 0x2000 */
        return charrom[((disp_start & 0x2000) >> 2) + (addr & 0x07FF)];
    }
#endif
}

//==============================================================================
// Video memory write.
//
// Note: If the 'standard' option is in affect then this code should behave like
// a standard machine. (extenram=0, attribram=0)
//
//   pass: UINT32 addr
//         UINT8 data
//         struct MemoryWriteByte *mem_s
// return: void
//==============================================================================
void vdu_vidmem_w (UINT32 addr, UINT8 data, struct MemoryWriteByte *mem_s)
{
// if (debuglvl > 0)
//    fprintf(debuglog, "vdu_vidmem_w at 0x%04x = 0x%02x\n", (unsigned int)addr, data);

#ifdef use_alpha_p
 int vb;

 if (! extendram)		// if the alpha+ graphics is not enabled
    vb = 0;
 else    
    vb = videobank;

 if (addr & 0x0800)    
    {
     if (! colourram)
        pcg_ram_ap[vb * 0x0800 + (addr & 0x07FF)] = data;
     else   
        {
#if 0
         printf("Colour RAM: %x\n", addr);
#endif
         colour_ram_ap[(vb & B8(00000011)) * 0x0800 + (addr & 0x07FF)] = data; 
        }
     z80mem[addr & 0xFFFF] = data;
     crtc_set_redraw();        
    }    
 else
    {
     if (! latchrom)
        {
         if (! attribram)
            screen_ram_ap[(vb & B8(00000011)) * 0x0800 + (addr & 0x07FF)] = data;
         else
            attrib_ram_ap[(vb & B8(00000011)) * 0x0800 + (addr & 0x07FF)] = data;
         z80mem[addr & 0xFFFF] = data;
         crtc_redraw_char(addr);
        }
    }
#else 
 if (addr & 0x0800)
    {
     pcgram[addr & 0x07FF] = data;
     z80mem[addr & 0xFFFF] = data;
     crtc_set_redraw();
    }
 else
    if (! latchrom)
       {
        videoram[addr & 0x07FF] = data;
        z80mem[addr & 0xFFFF] = data;
        crtc_redraw_char(addr);
       }
#endif       
}

//==============================================================================
// Read Port 0x08 - colour control
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			colour_cont
//==============================================================================
UINT16 vdu_colcont_r (UINT16 port, struct z80PortRead *port_s)
{
#ifdef use_debug_vdu_colcont_r
 printf("Colour control (0x08) read: %x\n", colour_cont); 
#endif
 return colour_cont;
}

//==============================================================================
// Write Port 0x08 - colour control
//
// Must switch in either the PCG or Colour RAM banks to 8000/8800 or F000/F800
// depending on the port data.
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_colcont_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
#ifdef use_debug_vdu_colcont_w
     printf("Colour control (0x08) write: %x\n", colour_cont); 
#endif
 if (usecolour)
    {
     colour_cont = data & B8(11111110);
     colourram = colour_cont & B8(01000000);

     if (x_colour_cont != colour_cont)
        {
         vdu_switchvideo_out();
         vdu_switchvideo_in();
        } 
    } 
}

//==============================================================================
// Read Port 0x09 - Colour wait off
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UNINT16			0
//==============================================================================
UINT16 vdu_colwait_r (UINT16 port, struct z80PortRead *port_s)
{
#ifdef use_debug_vdu_colwait_r
 printf("Colour wait off (0x09) read"); 
#endif
 return 0;
}

//==============================================================================
// Write Port 0x09 - Colour wait off
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_colwait_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
#ifdef use_debug_vdu_colwait_w
 printf("Colour wait off (0x09) write"); 
#endif
}

#ifdef use_alpha_p
//==============================================================================
// Read Port 0x1C - LV DATA 
//
//   pass: UINT16 port
//         struct z80PortRead *port_s
// return: UINT16			lv_dat
//==============================================================================
UINT16 vdu_lvdat_r (UINT16 port, struct z80PortRead *port_s)
{
#ifdef use_debug_vdu_lvdat_r
 printf("lvdat (0x1C) read: %x\n", lv_dat);
#endif

 return lv_dat;
}

//==============================================================================
// Write Port 0x1C - LV DATA
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_lvdat_w (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
#ifdef use_debug_vdu_lvdat_w
 printf("lvdat (0x1C) write: %x\n", lv_dat);
#endif

// uses default values of 0 for a standard machine if this option is enabled
 if (standard)
    return;

 lv_dat = data & B8(10111111);		// port (0x1c) value
 videobank = lv_dat & B8(00001111);	// bank bits 0-3
 attribram = lv_dat & B8(00010000);	// attribute RAM select
 extendram = lv_dat & B8(10000000); 	// extended graphics select

 if (x_lv_dat != lv_dat)
    {
     vdu_switchvideo_out();
     vdu_switchvideo_in();
    } 
}
#endif

//==============================================================================
// Port 0x0B - Character ROM select
//
//   pass: UINT16 port
//         UINT8 data
//         struct z80PortWrite *port_s
// return: void
//==============================================================================
void vdu_latchrom (UINT16 port, UINT8 data, struct z80PortWrite *port_s)
{
#ifdef use_debug_vdu_latchrom
 printf("latchrom (0x0B) write: %x\n", latchrom); 
#endif
 latchrom = data & B8(00000001);
}
